/** JOYCE: PCW emulator **************************************/
/**                                                         **/
/**                          Joyce.h                        **/
/**                                                         **/
/** This file contains declarations relevant to the         **/
/** emulation of the PCW. See Z80.h for #defines related to **/
/** Z80 emulation.                                          **/
/**                                                         **/
/*************************************************************/


#ifndef JOYCE_H		    /* Protect against multiple inclusion */

/* Standard libraries */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include <unistd.h>
#include <ctype.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/types.h>

/* PC specific */

#ifdef __MSDOS__
#include <dir.h>
#include <bios.h>
#include <conio.h>
#include <dos.h>
#include <dpmi.h>
#include <pc.h>
#include <go32.h>
#include <sys/movedata.h>
#include <sys/farptr.h>

/* GRX graphics library */

#include <grdriver.h>
#include <grx20.h>

/* SVASYNC serial port library */

#include <svasync.h>

#endif /* __MSDOS__ */

#ifdef UNIX
#include "Joyceunx.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

#include "Z80.h"            /* Z80 emulation declarations    */

#define BCDVERS 0x0136	/* v1.36 */

/* file names */
#define BOOT_SYS	"bootfile.emj"	/* boot image of CP/M system	*/
#define BOOT_CFG	"boot.cfg"	/* (v1.22) boot config file */
#define LOGFILE		"joyce.log"	/* Log file */

#define BOOT_DIR        "boot"		/* Subdirectory for boot files */
#define DISK_DIR	"disks"		/* Subdirectory for disk images */


#define PAGESIZE  0x4000L        /* Size of a RAM page            */
#define NORAM     0xFF           /* Byte to be returned from      */
                                 /* non-existing pages and ports  */
#define ZERORAM		0x00	 /* 00 byte for ram initialization */

/* JOYCE routines: */

void goodbye(int sig);

char *find_boot_file (char *fname);	/* Locate a boot file */
char *find_dsk_file  (char *fname);	/* Locate a disc file */
char *find_joyce_file(char *fname);	/* Locate settings file */

int init_joyce(int argc, char **argv);	/* Break argv[0] into   *
                                         * directory & filename */

void vga_setmode(byte mode); /* Switch the screen mode */
void vga_timer(void);	 /* The VGA driver receives this at 900Hz */
char InitScr(int type);  /* Initialise the screen        */
void UnitScr(void);	 /* De-initialise the screen     */
void DrawScr(void);	 /* Draw the PCW RAM to screen   */
void VidPort(unsigned char port, unsigned char value);
			 /* Emulate the video-RAM ports  */
void JoyceMenu(Z80 *R);  /* Implement the "options" menu */
void ExitConf(void);     /* Implement the "exit" menu    */
void screen_on(void);	 /* Switch screen display on */
void screen_off(void);	 /* Switch screen display off */
void screen_colour(int wr, int wg, int wb, int br, int bg, int bb);
void screen_switch(int mode);	/* Switch to 800x600 and back */

int mem_alloc(void);	 /* Allocate PCW memory - returns 256-2048 */
void mem_page(unsigned char port, unsigned char bank);
			 /* Page the memory */
void mem_lock(unsigned char byte);
			 /* Set reads & writes to the same bank */
 
int InitKbd(void);     		/* Initialise keyboard emulation */
void UnitKbd(void);    		/* Give the PC back its keyboard */
void PollKbd(char x, Z80 *R);	/* Transfer keyboard data to PCW RAM */
void KeyUp(void);      		/* Wait until all keys have been released */
int GetKeyNo(void);    		/* Get a keypress */
void KeySet(void);		/* Change in settings */
void KeyboardED(Z80 *R);	/* Keyboard ED xx operations */

void InitFDC(int FDCtype);	/* Initialise FDC emulation */
void UnitFDC(void);		/* Deinit FDC emulation */
void FDCport(byte port, byte value); /* FDC I/O */
byte FDCin(byte port);
void fdc_set_tcount(byte value);/* Set the FDC's TC line */
void fdc_mstart(void);		/* Motors on */
void fdc_mstop(void);		/* Motors off */
int  fdc_boot(int);		/* Load boot sector */
void img_eject(int unit);	/* "eject" a disk image */
int  img_new(int unit);		/* create new image */

void beep_on(void);		/* Beeper on */
void beep_off(void);		/* Beeper off */

void joyce_dprintf(char *, ...); /* Printf() to debugger output */
void dgets(char *S, int maxc);	/* Get string for debugger */
char dgetch(void);		/* Get character */
void dcls(void);		/* Clear debug console screen */

void goodbye(int);		/* Exit tidily */
void diewith(char *, int);	/* Exit tidily with message */

void InitPtr(void);		/* Initialise printer emulation */
void UnitPtr(void);		/* Deinitialise printer emulation */
void PtrSettings(void);		/* Printer settings may have changed */

void fid_com(Z80 *R);		/* (v1.21) COM port handler */
void set_com(void);		/* (v1.22) COM port reinitialise */
void com_init(void);		/* (v1.31) COM port startup */
void com_deinit(void);		/* (v1.31) COM port shutdown */

void boot_id(Z80 *R);		/* (v1.22) parse BOOT.CFG */

void popup_on (char *msg, int nsecs, int centred);   /* (v1.27) popup messages */
void popup_off(void);
void DrawPopup(void);
void Border3d(int x1, int y1, int x2, int y2, int raised); /* Draw 3D border */
void UnitMouse(void);
void mouse_reset(void);

#ifdef USE_SDL
void UpdateArea(int x, int y, int w, int h);
#else
#define UpdateArea(a,b,c,d)
#endif

/* Interface to CPMREDIR library */
void cpmredir_intercept(Z80 *R);

/* Two structures. One holds settings which should be persistent between
  sessions; the other holds state which should be saved in snapshots 
  (though as yet there is no snapshot support) */

#define SETTINGS_MAGIC "JOYCE settings\032\01"

typedef struct
{
	char magic[16];		/* Magic number */
	byte swap_alt;		/* Swap ALT and CTRL? */
	word cen_port;		/* 0-3 => LPT1-4 */
	byte screen_refresh;	/* Screen refresh rate, 144 on P75 */
	byte verbose;		/* Verbose output? */
	byte one_drive;		/* Set to nonzero for 1 floppy drive */
	byte boot_ro;		/* Write protect boot images? */
	signed char mousemode;	/* Mouse mode (1 = AMX 0 = Kempston -1=none) */
	byte grxscr;		/* Access screen using GRX or direct access? */
	char printer_file[160];	/* (v1.11) ASCII file for printer output */
	byte beep;		/* (v1.20) 0 => visual bell 1 => Audio bell */
	byte comport;		/* (v1.21) 0-3 => COM1 - COM4 */ 
	signed int mxscale;	/* (v1.23) mouse X scaling */
	signed int myscale;	/* (v1.23) mouse Y scaling */
	signed int mxspeed;	/* (v1.23) mouse X speed */
	signed int myspeed;	/* (v1.23) mouse Y speed */
	byte patchmouse;	/* (v1.27) Use mouse patch? */
	char com_in[160];	/* (v1.31) COM port redirect from file [Unix] */
	char com_out[160];	/* (v1.31) COM port redirect to   file [Unix] */
	byte swap_delete;	/* (v1.31) Swap Backspace & Delete     [Unix] */
/*

Additional emulator settings go here in future

*/
} JOYCE_SETTINGS;

/* JOYCE state - what needs to be saved in a snapshot */

#define STATE_MAGIC "JOYCE snapshot\032\01"

typedef struct
{
	char magic[16];
	
	Z80 Rg;				/* The Z80 CPU's registers */
	byte inf8;			/* Value returned by IN(0F8h) */	
	byte PSG[16];			/* AY-3-8912 registers */
	byte fdc_terminal_count;	/* Set to abort FDC transfer */
	byte fdc_int_mode;		/* FDC interrupt mode (INTs or NMIs?) */
	byte fdc_msr;			/* FDC main status register */
	long fdcint;			/* No. of instructions to FDC interrupt  */
	byte timerint;			/* Is a timer interrupt pending? */
	byte timeron;			/* Is the timer on? */
	short reset_flag;		/* Set to 1 to reset the Z80 */
	byte motor_on;			/* Drive motor running? */
	byte fdc_unit;			/* The unit the FDC is talking to */
	byte fdc_head;			/* Head within unit */
	byte model_type;		/* 0 => 8000 series, 1 => 9512 */
	byte mouse_patched;		/* (v1.27) mouse driver patched? */
	word md3_mx;			/* (v1.27) Address of mouse X for MD3 driver */
	word md3_my;			/* (v1.27) Address of mouse Y for MD3 driver */
	word md3_mf1;			/* (v1.27) Address of mouse flag 1 for MD3 driver */
	word md3_mf2;			/* (v1.27) Address of mouse flag 2 for MD3 driver */
	word sp_my;			/* (v1.27) Address of mouse X for Stop Press driver */
	word sp_mx;			/* (v1.27) Address of mouse Y for Stop Press driver */
	char popup[80];			/* (v1.27) Popup message */
	time_t popup_timeout;		/* (v1.27) When to hide popup */
	byte popup_at_bottom;		/* (v1.27) Popup centred or at bottom? */
	word mp_mx;			/* (v1.33) Address of mouse coords for MasterPaint */
} JOYCE_STATE;

#ifdef MAIN	/* Data storage classes - extern unless in main program */
#define EXTERN
#else
#define EXTERN extern
#endif

EXTERN int MouseButtons;		/* How many buttons has the mouse? */

EXTERN int DebugMode;			/* Screen in debug mode? */
EXTERN JOYCE_SETTINGS jset, jset1;	/* Emulation settings */	
EXTERN JOYCE_STATE jstate, jstate1;

EXTERN char bootfile[100];		/* The filename of the boot ROM */
EXTERN char logfile[100];		/* The filename of the log file */
EXTERN char *joycedir,*joycename;	/* The directory in which JOYCE is */
	 				/* located, and its name */

EXTERN FILE *fplog;			/* Log file */


EXTERN byte *PCWRAM;		/* The PCW's memory, all 2048k */
EXTERN byte *PCWRD[4];		/* Pointers to the current memory banks - reading */
EXTERN byte *PCWWR[4];		/* Pointers to the current memory banks - writing */
#ifdef __MSDOS__
EXTERN byte *Framebuf;		/* Back screen for display (800x600x256) */
#endif
EXTERN byte bank_lock[4];	/* Read & write to the same bank? */

/* FDC variables. A lot of them come in fours, but we only use the first
 * two (for exact PCW compatibility) */

EXTERN char *drive_filename[4];		/* Filenames for 4 disc images */
EXTERN FILE *drive_fp[4];		/* stdio files for the images */
EXTERN int drive_cylinder[4];		/* Head positions on the 4 images */
EXTERN int drive_readonly[4];		/* Read only? */
EXTERN signed char drive_fd[4];		/* Drives are files or real floppies? */

/* (v1.11) Printer output to text file */

EXTERN FILE *printer_fp;		/* stdio file for printer output */

/* (v1.21) COM port status */

EXTERN int comstat;			/* 0 if COM port available */

/* (v1.31) Requested by external agency to stop */

EXTERN int extern_stop;

/* (v1.33) Serial port provided? */

EXTERN int nsJoyce;

#ifdef __cplusplus
}
#endif


#define JOYCE_H 1

#endif /* ndef JOYCE_H */
