/** JOYCE PCW emulator for Linux/SDL *************************/
/**                                                         **/
/** JoyceSdlGfx.c: contains graphics primitives             **/
/**                                                         **/
/** Copyright (c) John Elliott 1996-1999                    **/
/**     You are not allowed to sell this software for       **/
/**     profit.                                             **/   
/*************************************************************/

/* JOYCE for DOS was written atop the GRX graphics library. JOYCE
 * for X uses SDL (there is a GRX for X, but for various reasons
 * I'm using SDL instead). This file emulates those GRX functions
 * that JOYCE uses and that aren't in SDL.
 *
 */

#include "Joyce.h"
#include "Joycevga.h"

static int clipx1, clipy1, clipx2, clipy2;

/* Graphics primitives */

static void GrDot(Uint8 *bits, Uint32 pixel, int bpp)
{
        switch(bpp)
        {
                case 1:
                *((Uint8 *)(bits)) = (Uint8)pixel;
                break;

                case 2:
                *((Uint16 *)(bits)) = (Uint16)pixel;
                break;

                case 3:
                { /* Format/endian independent */
                    Uint8 r, g, b;

                    r = (pixel>>context->format->Rshift)&0xFF;
                    g = (pixel>>context->format->Gshift)&0xFF;
                    b = (pixel>>context->format->Bshift)&0xFF;
                    *((bits)+context->format->Rshift/8) = r;
                    *((bits)+context->format->Gshift/8) = g;
                    *((bits)+context->format->Bshift/8) = b;
                }
                break;

                case 4:
                *((Uint32 *)(bits)) = (Uint32)pixel;
                break;
        }
}

static void GrDotRop(Uint8 *bits, Uint32 pixel, int bpp, int rop)
{
        switch(bpp)
        {
                case 1: switch(rop)
                {
			default: *((Uint8 *)(bits)) = (Uint8)pixel; break;
			case 1: *((Uint8 *)(bits)) &= (Uint8)pixel; break;
			case 2: *((Uint8 *)(bits)) |= (Uint8)pixel; break;
			case 3: *((Uint8 *)(bits)) ^= (Uint8)pixel; break;
		}
                break;

                case 2: switch(rop)
                {
                	default: *((Uint16 *)(bits))  = (Uint16)pixel; break;
                	case 1: *((Uint16 *)(bits)) &= (Uint16)pixel; break;
                	case 2: *((Uint16 *)(bits)) |= (Uint16)pixel; break;
                	case 3: *((Uint16 *)(bits)) ^= (Uint16)pixel; break;
                }
                break;

                case 3:
                { /* Format/endian independent */
                    Uint8 r, g, b;

                    r = (pixel>>context->format->Rshift)&0xFF;
                    g = (pixel>>context->format->Gshift)&0xFF;
                    b = (pixel>>context->format->Bshift)&0xFF;
                    switch(rop)
                    {
                      default:
                      *((bits)+context->format->Rshift/8) = r;
                      *((bits)+context->format->Gshift/8) = g;
                      *((bits)+context->format->Bshift/8) = b;
                      break;

                      case 1:
                      *((bits)+context->format->Rshift/8) &= r;
                      *((bits)+context->format->Gshift/8) &= g;
                      *((bits)+context->format->Bshift/8) &= b;
                      break;
                      
                      case 2:
                      *((bits)+context->format->Rshift/8) |= r;
                      *((bits)+context->format->Gshift/8) |= g;
                      *((bits)+context->format->Bshift/8) |= b;
                      break;
                      
                      case 3:
                      *((bits)+context->format->Rshift/8) ^= r;
                      *((bits)+context->format->Gshift/8) ^= g;
                      *((bits)+context->format->Bshift/8) ^= b;
                      break;
                    }
                }
                break;

                case 4: switch(rop)
		{
	                default: *((Uint32 *)(bits))  = (Uint32)pixel;
	                case 1: *((Uint32 *)(bits)) &= (Uint32)pixel;
	                case 2: *((Uint32 *)(bits)) |= (Uint32)pixel;
	                case 3: *((Uint32 *)(bits)) ^= (Uint32)pixel;
	        }
                break;
        }
}


static void GrXorDot(Uint8 *bits, int bpp)
{
	int n;
	for (n = 0; n < bpp; n++) bits[n] ^= 0xFF;
}



void GrPlotNC(int x, int y, int colour)
{
        int bpp;
        Uint32 pixel;
        Uint8 *bits;

        bpp = context->format->BytesPerPixel;

        pixel = SDL_MapRGB(context->format, colours[colour].r,
                                            colours[colour].g,
                                            colours[colour].b);

	bits = ((Uint8 *)context->pixels)+y*context->pitch+x*bpp;

	GrDot(bits, pixel, bpp);
}

int GrPixel(int x, int y)
{
        int bpp;
        Uint8 *bits;

        bpp = context->format->BytesPerPixel;
        bits = ((Uint8 *)context->pixels)+y*context->pitch+x*bpp;
        switch(bpp)
        {
                case 1:
		return *((Uint8 *)(bits));
                break;

                case 2:
                case 3:
		case 4:
                { /* Format/endian independent */
                    Uint8 r, g, b;

                    r = *((bits)+context->format->Rshift/8);
                    g = *((bits)+context->format->Gshift/8);
                    b = *((bits)+context->format->Bshift/8);
                    return colour_find(r,g,b);
                }
	}
	return 0;
}



void GrPlotRopNC(int x, int y, int colour, int rop)
{
        int bpp;
        Uint32 pixel;
        Uint8 *bits;

        bpp = context->format->BytesPerPixel;

        pixel = SDL_MapRGB(context->format, colours[colour].r,
                                            colours[colour].g,
                                            colours[colour].b);

	bits = ((Uint8 *)context->pixels)+y*context->pitch+x*bpp;

	GrDotRop(bits, pixel, bpp, rop);
}





void GrFilledBoxNC(int x1, int y1, int x2, int y2, int colour)
{
	int x, y, bpp;
	Uint32 pixel;
	Uint8 *bits;

        bpp = context->format->BytesPerPixel;

	pixel = SDL_MapRGB(context->format, colours[colour].r,
                                            colours[colour].g,
                                            colours[colour].b);

	for (y = y1; y <= y2; y++)
	{
		bits = ((Uint8 *)context->pixels)+ (y * context->pitch) + (x1 * bpp);
		for (x = x1; x <= x2; x++) 
		{	
			GrDot(bits, pixel, bpp);
			bits += bpp;
		}
	}
}

void xor_block(int y1, int x1, int h, int w)
{
        int x, y, bpp;
	Uint8 *bits;

        bpp = context->format->BytesPerPixel;

        for (y = y1; y < y1 + h; y++)
        {
                bits = ((Uint8 *)context->pixels)+ (y * context->pitch) + (x1 * bpp);
                for (x = x1; x <= x1 + w; x++)
                {
                        GrXorDot(bits, bpp);
                        bits += bpp;
                }
        }
}




void GrClearScreen(int index)
{
	GrFilledBoxNC(0, 0, 799, 599, index);
}

void GrBoxNC(int x1, int y1, int x2, int y2, int colour)
{
	GrHLineNC(x1, x2, y1, colour);
	GrHLineNC(x1, x2, y2, colour);
	GrVLineNC(x1, y1, y2, colour);
	GrVLineNC(x2, y1, y2, colour);
}




void GrHLineNC(int x1,int x2,int y, int c)
{
	int x,bpp;
	Uint32 pixel;
	Uint8 *bits;

        bpp = context->format->BytesPerPixel;

        pixel = SDL_MapRGB(context->format, colours[c].r,
                                            colours[c].g,
                                            colours[c].b);

        bits = ((Uint8 *)context->pixels)+y*context->pitch+x1*bpp;
        for (x = x1; x <= x2; x++)
        {
                GrDot(bits, pixel, bpp);
                bits += bpp;
        }
}




	
void GrVLineNC(int x,int y1,int y2, int c)
{
        int y, bpp;
        Uint32 pixel;
        Uint8 *bits;

        bpp = context->format->BytesPerPixel;

        pixel = SDL_MapRGB(context->format, colours[c].r,
                                            colours[c].g,
                                            colours[c].b);

        bits = ((Uint8 *)context->pixels)+y1*context->pitch+x*bpp;
        for (y = y1; y <= y2; y++)
        {
                GrDot(bits, pixel, bpp);                
                bits += context->pitch;
        }
}

void GrHLine(int x, int y1, int y2, int c)
{
	if (x  < clipx1)  x = clipx1;
	if (y1 < clipy1) y1 = clipy1;
	if (y2 < clipy1) y2 = clipy1;
	if (x >  clipx2)  x = clipx2;
	if (y1 > clipy2) y1 = clipy2;
	if (y2 > clipy2) y2 = clipy2;
	if (y1 > y2) { int t; t = y1; y1 = y2; y2 = t; }
	GrHLineNC(x, y1, y2, c);
}

void GrVLine(int x1, int x2, int y, int c)
{
        if (y  < clipy1)  y = clipy1;
        if (x1 < clipx1) x1 = clipx1;
        if (x2 < clipx1) x2 = clipx1;
        if (y  > clipy2)  y = clipy2;
        if (x1 > clipy2) x1 = clipx2;
        if (x2 > clipy2) x2 = clipx2; 
        if (x1 > x2) { int t; t = x1; x1 = x2; x2 = t; }
        GrVLineNC(x1, x2, y, c);
}



void GrClearClipBox(int col)
{
        GrFilledBoxNC(clipx1, clipy1, clipx2, clipx2, col);
	UpdateArea(clipx1, clipy1, clipx2, clipy2);
}


void GrSetClipBox(int x1, int y1, int x2, int y2)
{
	clipx1 = x1; clipx2 = x2;
	clipy1 = y1; clipy2 = y2;

	if (clipx1 < 0)   clipx1 = 0;
	if (clipx2 < 0)   clipx2 = 0;
	if (clipx1 > 799) clipx1 = 799;
	if (clipx2 > 799) clipx2 = 799;

        if (clipy1 < 0)   clipy1 = 0;
        if (clipy2 < 0)   clipy2 = 0;
        if (clipy1 > 599) clipy1 = 599;
        if (clipy2 > 599) clipy2 = 599;

	if (clipx2 < clipx1) { int t; t = clipx1; clipx1 = clipx2; clipx2 = t; }
	if (clipy2 < clipy1) { int t; t = clipy1; clipy1 = clipy2; clipy2 = t; }
}


void GrFilledBox(int x1, int y1, int x2, int y2, int colour)
{
        if (x1 < clipx1) x1 = clipx1;
        if (x2 < clipx1) x2 = clipx1;
        if (x1 > clipx2) x1 = clipx2;
        if (x2 > clipx2) x2 = clipx2;

        if (y1 < clipy1) y1 = clipy1;
        if (y2 < clipy1) y2 = clipy1;
        if (y1 > clipy2) y1 = clipy2;
        if (y2 > clipy2) y2 = clipy2;

        if (x2 < x1) { int t; t = x1; x1 = x2; x2 = t; }
        if (y2 < y1) { int t; t = y1; y1 = y2; y2 = t; }

	GrFilledBoxNC(x1, y1, x2, y2, colour);
}





void GrPlotRop(int x1, int y1, int colour, int rop)
{
        if (x1 < clipx1) x1 = clipx1;
        if (x1 > clipx2) x1 = clipx2;

        if (y1 < clipy1) y1 = clipy1;
        if (y1 > clipy2) y1 = clipy2;

	GrPlotRopNC(x1, y1, colour, rop);
}



void GrPlot(int x1, int y1, int colour)
{
        if (x1 < clipx1) x1 = clipx1;
        if (x1 > clipx2) x1 = clipx2;

        if (y1 < clipy1) y1 = clipy1;
        if (y1 > clipy2) y1 = clipy2;

        GrPlotNC(x1, y1, colour);
}

/* Draw a 3d border around something */

void Border3d(int x1, int y1, int x2, int y2, int raised) /* v1.27: not static */
{
        GrHLine(x1,x2,y1, raised ? shadow_hi : shadow_lo);
        GrVLine(x1,y1,y2, raised ? shadow_hi : shadow_lo);
        GrVLine(x2,y1,y2, raised ? shadow_lo : shadow_hi);
        GrHLine(x1,x2,y2, raised ? shadow_lo : shadow_hi);
}



void UpdateArea(int x, int y, int w, int h)
{
	SDL_UpdateRect(screen, x,y,w,h);
}

void SdlScrBlit(int dx, int dy, int sx, int sy, int sx2, int sy2)
{
	Uint8 *sbits, *dbits;
	int bpp = screen->format->BytesPerPixel;
	int lpi = screen->pitch;
	int y,w = (sx2 - sx + 1) * bpp;

	if (SDL_LockSurface(screen) < 0 ) return;

	if (sy < dy && dy < sy2) for (y = (sy2-sy); y > 0; y--)
        {
                sbits = ((Uint8 *)screen->pixels) + ((y+sy) * lpi) + (sx * bpp);
                dbits = ((Uint8 *)screen->pixels) + ((y+dy) * lpi) + (dx * bpp);

                memmove(dbits, sbits, w);
        }

        else for (y = 0; y < (sy2-sy); y++)
        {
                sbits = ((Uint8 *)screen->pixels) + ((y+sy) * lpi) + (sx * bpp);
		dbits = ((Uint8 *)screen->pixels) + ((y+dy) * lpi) + (dx * bpp);

		memmove(dbits, sbits, w);
	}

	SDL_UnlockSurface(screen);
}

__inline__ int sgn(int x)
{
	/* Beautiful, isn't it? */
	return (!x ? 0 : (( x < 0 ) ? -1 : 1));
}


/* This line-drawing algorithm comes straight from the ZX81 manual, I kid
   you not. The comments are transcribed exactly, and the ZX81 has no
   lower-case letters; hence the shouting. */

void GrLineRop(int a, int b, int c, int d, int colour, int rop)
{
	int u   = c - a;	/* U SHOWS HOW MANY STEPS ALONG WE NEED TO GO */
	int v   = d - b;	/* V SHOWS HOW MANY STEPS UP */
	int d1x = sgn(u);	/* (D1X,D1Y) IS A SINGLE STEP IN A */
	int d1y = sgn(v);	/*           DIAGONAL DIRECTION    */ 
	int d2x = d1x;
	int d2y = 0;		/* (D2X,D2Y) IS A SINGLE STEP LEFT OR RIGHT */
	int m   = abs(u);
	int n   = abs(v);
	int s, i;

	if (m <= n)
	{
		d2x = 0;
		d2y = d1y;	/* NOW (D2X,D2Y) IS A SINGLE STEP UP OR DOWN */
	
		m = abs(v);
		n = abs(u);	
	}			/* M IS THE LARGER OF ABS U & ABS V, N IS */
				/* THE SMALLER */
	s = m / 2;		/* WE WANT TO MOVE FROM (A,B) TO (C,D) IN */
				/* M STEPS USING N UP-DOWN OR RIGHT-LEFT  */
				/* STEPS D2, & M-N DIAGONAL STEPS D1, */
				/* DISTRIBUTED AS EVENLY AS POSSIBLE */
	for (i = 0; i < m; i++)
	{
		GrPlotRop(a,b,colour,rop);
		s += n;
		if (s >= m)
		{
			s -= m;
			a += d1x;
			b += d1y;	/* A DIAGONAL STEP */
		}
		else
		{
			a += d2x;
			b += d2y;	/* AN UP-DOWN OR LEFT-RIGHT STEP */
		}
	}	
}



