/** JOYCE PCW emulator for DJGPP *****************************/
/**                                                         **/
/** JOYCEMOU.C: Emulates the PCW mouse via a Microsoft or   **/
/**            MouseSystems mouse                           **/
/**                                                         **/
/** Copyright (c) John Elliott 1996-1999                    **/
/**     You are not allowed to sell this software for       **/
/**     profit.                                             **/   
/*************************************************************/

#include "Joyce.h"
#include "Joycemou.h"
#include "inline.h"

/* New in v1.23: better mouse handling.

[Terminology: a 'mickey' is the basic unit of mouse movement]

  The pre-v1.23 handling in AMX mode was very simple:

      PCW mickeys = (PC mickeys & 0x0F)

  This meant that any move above 15 mickeys would become a move by a random
 number below 15. In turn, this meant that large moves were jerky and 
 inaccurate.

  New behaviour: 
   * The PC mickey count can be scaled (eg 4 PC mickeys : 1 PCW mickey). 
    This improves accuracy on small moves. The scale factors are jset.mxscale
    and jset.myscale.
   * The equation is now: PCW mickeys = min ((PC mickeys / scale), speed)
    where the speed is jset.mxspeed / jset.myspeed. This means that if the
    PC mickey count is over (say) 15, 15 is returned. Reducing the value
    of 'speed' reduces the number of PCW mickeys that can be returned by
    one call, making motion smoother if slower.
   * If these transformations mean that not all the PC mickeys have been
    "used up", then the unused ones are kept for next time. This improves
    long-distance navigation as the big PC move is translated into separate
    PCW moves rather than truncated.

  The Kempston mouse uses 8-bit mickey counts rather than 4-bit so is less
likely to be affected by the original problem.

 */

#define min(a,b) (a < b ? a : b)
#define max(a,b) (a > b ? a : b)

/* These figures define the mouse ballistics */

/* Values as per JOYCE v1.22
 *
 * jset.mxscale = 1; jset.myscale = 1;
 * jset.mxspeed  = 15, jset.myspeed  = 15; 
 *
 */

/* These make it smoother but slower in MD3 
 *
 * jset.mxscale = 4, jset.myscale = 2;
 * jset.mxspeed  = 7, jset.myspeed  = 7; 
 *
 */


void MouseRead(short *dy, short *dx)    /* Read the no. of mickeys moved */
{
	static int ox, oy, x, y;

	SDL_GetMouseState(&x, &y);

        *dx += (x-ox);
        *dy += (y-oy);

	ox = x;	
	oy = y;
}


void MouseBtn(word *btn)        /* Check for buttons */
{
	int x, y;

	*btn = SDL_GetMouseState(&x, &y);
}




/* (v1.23) mouse scaling options (PC mickeys : PCW mickeys)
 */

static signed short scale_x(signed short x)
{
	if (x < 0) x = (-x);
	return min((x / jset.mxscale), jset.mxspeed);
}


static signed short scale_y(signed short y)
{
	if (y < 0) y = (-y);
	return min((y / jset.myscale), jset.myspeed);
}



void InitMouse(void)	/* Initialise the mouse driver */
{
	MouseButtons = 3;	/* On X, we assume 3 buttons. */
}


/* (v1.27)  New mouse capture code

              The idea of this code is to bypass all that tedious mucking 
            about with I/O ports and mickeys, and to pass mouse coordinates
            directly into the program.
 
               The problem is: there is no standard mouse driver interface.
             What we have to do is trap a mouse read, and find out where the
             program keeps its mouse driver code. We then patch it.
*/ 

/* According to the PCW Plus Tips Collection:
 * AMX:
 *
 * IN A0 gives vertical movement -   low  nibble = no. of moves up
 *                                   high nibble = no. of moves down
 * IN A1 gives horizontal movement - low  nibble = no. of moves right
 *                                   high nibble = no. of moves left
 * IN A2 gives button state in bottom 3 bits
 */


void mouse_patch(Z80 *R)	/* Called when our patch is triggered */
{
	Uint8  btn;
	int col, row;

	btn = SDL_GetMouseState(&col, &row);

/*** This code (used in DOS) maps the whole screen area to the PCW mouse; so that the real mouse has to be
    at the edge of the physical screen when the PCW mouse is at the edge of its border. In an X window
    (especially when focus follows the mouse) this is too confusing.

	row = (256L * row) / 600;
	col = (720L * col) / 800;
*/

	row /= 2;
	if (col < 40) col = 0; else col -= 40;	/* Limit to within the border area */
	if (row < 22) row = 0; else row -= 22;

	if (col > 719) col = 719;
	if (row > 255) row = 255;

	if (jstate.md3_mx)	/* MD3 patch active? */
	{
		row = 255 - row;
		if (row == RdWord(jstate.md3_my) && (col == RdWord(jstate.md3_mx))) 
		{
			R->AF.B.h = 0;
			return;
		}
		WrWord(jstate.md3_mx,  col);
		WrWord(jstate.md3_my,  row);
		WrZ80(jstate.md3_mf1, 0xFF);
		WrZ80(jstate.md3_mf2, 0xFF);
		R->AF.B.h = 1;
	}
	if (jstate.sp_mx)
	{
		WrWord(jstate.sp_mx, col);
		WrWord(jstate.sp_my, row);
	}
        if (jstate.mp_mx)
        {
                WrWord(jstate.mp_mx, col);
                WrWord(jstate.mp_mx+2, row);
	}
}

static int zmemcmp(word addr, byte *compare, int count)
{
	int n;

	for (n = 0; n < count; n++) if (RdZ80(addr++) != compare[n]) return -1;
	return 0;
}

static void zmemput(word addr, byte *replace, int count)
{
	int n;
	for (n = 0; n < count; n++) WrZ80(addr++, replace[n]);
}


/* Called the first time a mouse port is read. Checks if the mouse driver
  is anything we recognise */

void check_for_patch(void)
{
	int    patch_add = jstate.Rg.PC.W - 2;	/* Address of calling code */

	static byte md3_match1[5] = {0xDB, 0xA2, 0xF6, 0xF8, 0xC9};
	static byte md3_match2[6] = {0xCD, 0x00, 0x00, 0x28, 0x0F, 0x2A};
	static byte md3_patch2[6] = {0x3E, 0xFD, 0xED, 0xFE, 0xA7, 0xC9};
	static byte sp_match[9]   = {0xED, 0x78, 0xFE, 0x10, 0x20, 
                                     0x0E, 0x01, 0x80, 0x81};
	static byte sp_patch[7]   = {0x3E, 0xFD, 0xED, 0xFE, 0xC3, 0x00, 0x00};
        static byte mp_match1[7]  = {0xDB, 0xA2, 0x3C, 0x28, 0x49, 0x3D, 0x21};
        static byte mp_match2[14] = {0xDB, 0xD0, 0x47, 0xDB, 0xD1, 0x4F, 0xDB,
                                     0xD4, 0x57, 0xA0, 0xA1, 0x3C, 0x28, 0x50};

	jstate.mouse_patched = -1;
        /* Check for MasterPaint code */
        if (jstate.mouse_patched == 0) jstate.mouse_patched = 0x80;

        /* Check for MasterPaint code. The first check triggers the patch
          mechanism, but we're not ready to do the patch yet. So we put
          it into patch mode 2 (watch permanently). The second check is
          called when we are ready to do the patch. */

        if (!zmemcmp(patch_add, mp_match1, 7))
        {
                jstate.mouse_patched = 0x81;
                popup_on("Detected MasterPaint...", 0, 1);
        }
        if (!zmemcmp(patch_add, mp_match2, 14))
        {
                jstate.mp_mx = RdWord(patch_add + 0x69);
                sp_patch[4] = 0xC9;
                zmemput(patch_add + 0x68, sp_patch, 5);
                sp_patch[4] = 0xC3;
                jstate.mouse_patched = 3;
                popup_on("Applied MasterPaint mouse patch", 4, 1);
        }

	/* Check for md3 code */
	if (!zmemcmp(patch_add, md3_match1, 5)) 
	{
		md3_match2[1] = (patch_add - 0x20) & 0xFF;
		md3_match2[2] = (patch_add - 0x20) >> 8;
		if (!zmemcmp(patch_add - 0x254, md3_match2, 6))
		{	
			zmemput(patch_add - 0x254, md3_patch2, 6);

			jstate.md3_my  = RdWord(patch_add - 0x24E);
			jstate.md3_mx  = RdWord(patch_add - 0x23B);
			jstate.md3_mf1 = RdWord(patch_add - 0x228);
			jstate.md3_mf2 = RdWord(patch_add - 0x225);

			popup_on("Applied MD3 mouse patch", 4, 1);
			jstate.mouse_patched = 1;
		}
	}
	/* Check if we can recognise the Stop Press mouse code; and if we can,
           hook it. */
	if (!zmemcmp(patch_add, sp_match, 9))
	{
		sp_patch[5] = (patch_add + 0x103) & 0xFF;
		sp_patch[6] = (patch_add + 0x103) >> 8;
		zmemput(patch_add + 0x88, sp_patch, 7);
	
		jstate.sp_mx = RdWord(patch_add + 0x25);
		jstate.sp_my = RdWord(patch_add + 0x2b);		
		popup_on("Applied Stop Press mouse patch", 3, 1);
		jstate.mouse_patched = 2;
	}

	if (jstate.mouse_patched > 0)	/* Remove X's own cursor */
	{
		static int x, y;

		SDL_GetMouseState(&x, &y);

		tracking = 1;
		track_mouse(x, y);
	}

}

/* End v1.27 code */

byte MouseIn(byte port)	/* Deal with input from mouse port */
{
	byte rv=0xFF;
	static signed short y=0,x=0;
	word btn;

	if (jset.mousemode < 0 || MouseButtons < 0) return 0xFF;

	/* (v1.27) MD3 patch code */
	if ((jstate.mouse_patched == 0 || jstate.mouse_patched == 0x81)
             && jset.patchmouse) check_for_patch();

	if (jset.mousemode == 1) /* AMX */ switch(port)
	{
		case 0xA0:	/* (v1.23) Read & scale vertical movement */
		MouseRead(&y,&x);
		if (y <= 0) 
		{	
			rv =  scale_y(y)       & 0x0F;
			y  += min(-y, jset.myspeed);
		}
		else
		{
			rv = (scale_y(y) << 4) & 0xF0;
			y -= min(y, jset.myspeed);
		}
		//y = 0;	/* reset Y count */
		return rv;

		case 0xA1:	/* Read horizontal movement */
		MouseRead(&y,&x);
		if (x >= 0) 
		{
			rv =  scale_x(x)       & 0x0F;
			x -=  min(x,jset.mxspeed);
		}
		else
		{
			rv = (scale_x(x) << 4) & 0xF0;
			x += min(-x, jset.mxspeed);
		}
		//x = 0;	/* reset X count */
		return rv;
		
		case 0xA2:
		MouseBtn(&btn);
		rv = 0xFF;
		if (btn & 1) rv &= 0xFE;	/* L */
		if (btn & 2) rv &= 0xFD;	/* M */
		if (btn & 4) rv &= 0xFB;	/* R */		
		return rv;
	
		case 0xA3:
		return 0;
	}

	else  /* Kempston */ switch(port)
	{
		case 0xD0:	/* Read horizontal movement */
		MouseRead(&y,&x);
		return (x >> 3);

		case 0xD1:	/* Read vertical movement */
		MouseRead(&y,&x);
		return (-(y >> 3));

		case 0xD2:	/* ? */
//		MouseRead(&y,&x);
		return 0;

		case 0xD3:	/* ? */
//		MouseRead(&y,&x);
		return 0;
		
		case 0xD4:
		MouseBtn(&btn);
		rv = 0xFF;
		if (btn & 1) rv &= 0xFD;	/* L */
		if (btn & 2) rv &= 0xFE;	/* R */
		if (btn & 4) rv &= 0xFE;	/* M */
		return rv;
	}
	return 0xFF;
}


void UnitMouse(void)
{

}


void mouse_reset(void)
{
	SDL_ShowCursor(1);
	tracking = 0;
}


void track_mouse(Uint16 x, Uint16 y)
{
	static int sc = -1;

	if (!tracking)
	{
        	SDL_ShowCursor(1);
		return;
	}

	if (x < 40 || x >= 760 || y < 44 || y >= 556)
	{
		if (sc != 1) SDL_ShowCursor(1);
		sc = 1;
	}
	else
	{
		if (sc != 0) SDL_ShowCursor(0);
		sc = 0;
	}
}

