/** JOYCE PCW emulator for DJGPP *****************************/
/**                                                         **/
/** JOYCEDOS - interface between Z80 programs and the host  **/
/**            OS                                           **/
/**                                                         **/
/** A vaguely BDOS-like interface - C=function and          **/
/** DE=parameter. It can also take values in B and HL.      **/
/**                                                         **/
/** Returns HL=0 on failure, other values on success.       **/
/**                                                         **/
/** Copyright (c) John Elliott 1996-2000                    **/
/**     You are not allowed to sell this software for       **/
/**     profit.                                             **/   
/*************************************************************/

#include "Joyce.h"
#include "Joycedos.h"
#include "inline.h"

static FILE *handles[16];	/* Can open up to 15 files; handles[0] is 
				 * reserved */
static char pathname[162];	/* DOS pathname passed by Z80 program */

#define BCDIFY(x) (((x % 10)+16*(x/10)) & 0xFF)


void jdos_time(Z80 *R, int secflag)	/* Read the time into buffer at DE */
{
	time_t secs;			/* If secflag=1, include seconds */
	struct tm *ltm;
	int day, hour, min, sec, lyear;

	time(&secs);
					/* << v1.36: Use local time, 
					 *    rather than UTC. */
	ltm  = localtime(&secs);

	/* Work out the day count. */
	lyear = (ltm->tm_year - 76);	/* Years since 1976 */

	day = (lyear / 4) * 1461;	/* No. of days to start of 
					 * last leapyear */
	lyear = (lyear % 4);
	if (lyear)    { day += 366; --lyear; }
	while (lyear) { day += 365; --lyear; }
	/* Day = day count from 1976 to start of this year */
	day -= (365 + 366 - 1);	
	/* The extra 1 is because the CP/M day count is 1 based */
	day += ltm->tm_yday;

	hour = ltm->tm_hour;
	min  = ltm->tm_min;
	sec  = ltm->tm_sec;	
				/* >> v1.36 */
	if (!R->DE.W) return;
	WrZ80(R->DE.W, day & 0xFF);
	WrZ80(R->DE.W + 1, (day >> 8) & 0xFF);
	WrZ80(R->DE.W + 2, BCDIFY(hour));
	WrZ80(R->DE.W + 3, BCDIFY(min));
	if (secflag) WrZ80(R->DE.W + 4, BCDIFY(sec));
	R->AF.B.h = BCDIFY(sec);

}
#undef BCDIFY


void txt2fcb(word addr, char *s)	/* Parse ASCII filename to FCB */
{
	char curch;
	int fcnt = 0;

	while(fcnt < 11) switch(*s)
	{
		case 0:	while (fcnt < 11)
			{
				WrZ80(addr + fcnt, ' ');
				fcnt++;
			}
			return;

		case '.': while (fcnt < 8)
			{
				WrZ80(addr + fcnt, ' ');
				fcnt++;			
			}
			++s;
			break;

		case ':': 
		case '/':	
		case '\\':
			fcnt = 0;
			++s;
			break;

		default:
			/* v1.11: Deal with W95 or Unix filenames which
			 * may have lowercase letters in them */
			curch = (*s) & 0x7F; 
			if (islower(curch)) curch = toupper(curch);
			WrZ80(addr + fcnt, curch);
			++s;
			++fcnt;
			break;

	}
}

#ifdef __MSDOS__

#define FIND_ATTR FA_RDONLY | FA_HIDDEN | FA_ARCH

word jdos_wildex(Z80 *R)	/* Expand DOS wildcard */
{
	struct ffblk fblk;
	int m = 0;
	char c;

	do {
		c = RdZ80(R->DE.W + m);	/* Extract pathname */
		pathname[m] = c;
		++m;
	} while (c);

	if (findfirst(pathname, &fblk, FIND_ATTR)) return 0;

	m = R->BC.B.h;
	while (m > 0)
	{
		if (findnext(&fblk)) return 0;
		--m;
	}
	
	/* Substitute filename into pathname */

	m = strlen(pathname) - 1;
	if (m < 0) return 0;

	while (pathname[m] != '\\' && 
	       pathname[m] != '/'  && 
               pathname[m] != ':'  && m >= 0) --m;	
	
	++m;
	strcpy(pathname + m, fblk.ff_name);	/* Generated pathname */

	txt2fcb(R->HL.W, fblk.ff_name);
	m = (-1);
	do {
		++m;
		WrZ80(R->HL.W + m + 11, pathname[m]);
	} while (pathname[m]);
	return 1;
}
#undef FIND_ATTR

#endif /* __MSDOS__ */


word jdos_creat(Z80 *R)	/* Create a file, DE->name */
{
	int n, m;
	char c;

	for (n = 1; n <16; n++)	/* Find a free handle */
	{
		if (!handles[n]) break;
	}
	if (n == 16) return 0;	/* No free handles */
	
	m = 0;
	do {
		c = RdZ80(R->DE.W + m);	/* Extract pathname */
		pathname[m] = c;
		++m;
	} while (c);

	handles[n] = fopen(pathname, "w+b");	/* Create the file */

	if (!handles[n]) return 0;

	return n;
}

word jdos_open(Z80 *R)	/* Open a file, DE->name */
{
	int n, m;
	char c;

	for (n = 1; n <16; n++)	/* Find a free handle */
	{
		if (!handles[n]) break;
	}
	if (n == 16) return 0;	/* No free handles */
	
	m = 0;
	do {
		c = RdZ80(R->DE.W + m);	/* Extract pathname */
		pathname[m] = c;
		++m;
	} while (c);

	handles[n] = fopen(pathname, "r+b");	/* Open the file */

	if (!handles[n]) /* Read/write failed. Try read-only */
		handles[n] = fopen(pathname, "rb");
	if (!handles[n]) return 0;

	return n;
}



word jdos_putc(Z80 *R)	/* Write byte to file */
{
	int n = R->DE.W;

	if (!handles[n]) return 0;	/* Not open */

	if (fputc(R->BC.B.h, handles[n]) == EOF) return 0;

	return 1;
}


word jdos_getc(Z80 *R)	/* Read a byte from a file into B */
{
	int n = R->DE.W;
	int m;

	if (!handles[n]) return 0;	/* Not open */

	m = fgetc(handles[n]);
	if (m == EOF) return 0;		/* EOF */
	R->BC.B.h = (m & 0xFF);
	return 1;
}

word jdos_close(Z80 *R)	/* Close file */
{
	int n = R->DE.W;
	int m;

	if (!handles[n]) return 0;

	m = fclose(handles[n]);

	handles[n] = NULL;

	return(!m);	
}


void jdos(Z80 *R)
{
	switch(R->BC.B.l)
	{
		case 0x0F:	/* Open */
		R->HL.W = jdos_open(R);
		break;

		case 0x10:	/* Close */
		R->HL.W = jdos_close(R);
		break;

		case 0x14:	/* Get byte from stream */
		R->HL.W = jdos_getc(R);
		break;

		case 0x15:	/* Write byte to stream */
		R->HL.W = jdos_putc(R);
		break;

		case 0x16:
		R->HL.W = jdos_creat(R); /* Create file */
		break;

		case 105:
		jdos_time(R, 0);	/* Get the time */
		break;


		case 0x91:		/* Wildcard expansion */
		R->HL.W = jdos_wildex(R);
		break;

		default:
		joyce_dprintf("Unknown call to API function 10: "
			"C=%02x DE=%04x PC=%04x\n",
			R->BC.B.l, R->DE.W, R->PC.W);
		break;
	}
}

