/** JOYCE PCW emulator for DJGPP *****************************/
/**                                                         **/
/** JOYCEDSK: Access to PC floppies using the PC BIOS       **/
/**                                                         **/
/** Copyright (c) John Elliott 1996-1998                    **/
/**     You are not allowed to sell this software for       **/
/**     profit.                                             **/   
/*************************************************************/

#include "Joyce.h"
#include "Joycedsk.h"
#include "inline.h"

/* Floppy access is conducted using the PC BIOS and the FID interface.
 *
 * This is much easier (for me) than hardware-level emulation
 *
 * - Code to be called from a FID driver (JOYCEDRV.FID)
 *
 */
static byte dskparam[11];
static byte dskinfo[8192];		/* Disc specification from the boot sector */
static byte secbuf[8192];
static byte logged_in = 0;


/* Convert track number to cylinder+head */

static void calc_track(int track, int *cyl, int *head)
{
	switch(dskinfo[1] & 3)
	{
		case 0: *head = 0;
			*cyl  = track;
			break;

		default:
		case 1: *head = track & 1;
			*cyl  = (track >> 1);
			break;

		case 2: if (track > dskinfo[2])
			{
				*head = 1;
				*cyl  = 2*dskinfo[2] - track;
			}
			else
			{
				*head = 0;
				*cyl  = track;
			}
			break;
	}
}

static byte map_err(int err)
{
	switch(err)
	{
		case 3: return 2;	/* Disc is read only */
		case 6: return 0xFF;	/* Media changed */
	}
	return 1;
}


word to_al0(word blocks)
{
	word al0 = 0;

	while (blocks--) al0 = (al0 >> 1) | 0x8000;
	return al0;
}


byte tomask(byte b)
{
	byte mask = 0;

	while (b--) mask = (mask << 1) | 1;
	return mask;
}


word sizefrom(byte b)
{
	word W = 128;
	
	while (b--) W *= 2;

	return W;
}


void make_dpb(word dpb, byte *format, int is_xdpb)
{
	word BLOCK_SIZE, OFF, NUM_TRACKS, NUM_SECTORS, DIR_BLOCKS,
	     SEC_SIZE;

	word SPT, BSH, BLM, DSM, DRM, AL0, AL1, EXM, CKS, PSH, PHM;

	BLOCK_SIZE  = sizefrom(format[6]);
	OFF         = format[5];
	NUM_TRACKS  = format[2]; if (format[1] & 3) NUM_TRACKS *= 2;
	NUM_SECTORS = format[3];
	DIR_BLOCKS  = format[7];
	SEC_SIZE    = sizefrom(format[4]);

	SPT = NUM_SECTORS * SEC_SIZE  / 128; 	
	BSH = format[6];	
	BLM = tomask(format[6]);
	DSM = ((NUM_TRACKS - OFF) * NUM_SECTORS * SEC_SIZE / BLOCK_SIZE) - 1;
	DRM = (DIR_BLOCKS * BLOCK_SIZE / 32) - 1;
	AL0 = (to_al0(DIR_BLOCKS) >> 8);
	AL1 = (to_al0(DIR_BLOCKS) & 0xFF);
	if (DSM < 256) EXM = (BLOCK_SIZE >> 10) - 1;
	else	       EXM = (BLOCK_SIZE >> 11) - 1;
	CKS = (DRM + 1) / 4;
	PSH = format[4];
	PHM = tomask(format[4]);

	WrWord(dpb,     SPT);
	WrZ80 (dpb + 2, BSH);
	WrZ80 (dpb + 3, BLM);
	WrZ80 (dpb + 4, EXM);
	WrWord(dpb + 5, DSM);
	WrWord(dpb + 7, DRM);
	WrZ80 (dpb + 9, AL0);
	WrZ80 (dpb +10, AL1);
	WrWord(dpb +11, CKS);
	WrWord(dpb +13, OFF);
	WrZ80 (dpb +15, PSH);
	WrZ80 (dpb +16, PHM);
}



/*** Not possible in UNIX. Or rather it is - with setfdprm - but I am _not_
 *   building that into JOYCE.

void get_dskparam(byte *param)
{
	dosmemget(get_int1E(), 11, param);	
}

void put_dskparam(byte *param)
{
	dosmemput(param, 11, get_int1E());
	biosdisk(0, 0, 0, 0, 0, 0, NULL);	
}
***/


void setup_our_param()
{
	dskparam[3] = dskinfo[4];
	dskparam[4] = dskinfo[3];
	dskparam[5] = dskinfo[8];
	dskparam[7] = dskinfo[9];
	dskparam[8] = 0xE5;
}


void setup_dflt_param(void)
{
	dskparam[3] = 2;	/* 512 bytes / sector */
	dskparam[4] = 9;	/* 9 sectors / track */
	dskparam[5] = 0x2A;	/* R/W gap */
	dskparam[8] = 0xE5;	/* Format filler */
}


int bios_retry(int cmd, int drive, int head, int track, int sector, 
		int nsects, void *buffer)
{
	char drvname[50];
	FILE *fi;
	long offs;
	int secsz, n, rv = 0;

	sprintf(drvname, "/dev/fd%d", drive);
	fi = fopen(drvname, "r+b");

	/* Is read-only access possible? */
	if (!fi) fi = fopen(drvname, "rb");

	if (!fi) return 1;

	/* sector - 1, because sectors are 1 based */
	offs = (dskparam[4] * (track * 2 + head)) + sector - 1;

	secsz = 128;
	for (n = 0; n < dskparam[3]; n++) secsz *= 2;	

	if (fseek(fi, offs * secsz, SEEK_SET))
	{
		fclose(fi);
		return 1;
	}

	if (cmd == 2) rv = fread (buffer, secsz, nsects, fi);
	if (cmd == 3) rv = fwrite(buffer, secsz, nsects, fi);

	if (rv == nsects) rv = 0; else rv = 1;

	fclose(fi);
	return rv;
}


void fid_fd_login(int drive, Z80 *R)
{
	int err;
	setup_dflt_param();

	err = bios_retry(2, drive, 0, 0, 1, 1, dskinfo);

	if (err)  
	{
		R->AF.B.h = map_err(err);
		R->BC.B.h = err;
		R->AF.B.l &= ~C_FLAG;
		R->AF.B.l &= ~Z_FLAG;
		return;
	}
	logged_in = 1;
	
	/* v1.31: Support PCW16 extended formats */ 
	if (dskinfo[0] == 0xEB || dskinfo[0] == 0xE9)
	{
		if (!memcmp(dskinfo + 0x2B, "CP/M", 4) &&
		    !memcmp(dskinfo + 0x33, "DSK",  3) &&
                    !memcmp(dskinfo + 0x7C, "CP/M", 4))
		{
			memcpy(dskinfo + 0, dskinfo + 0x80, 16);
		}
	}	/* end v1.31 change */

	make_dpb(R->IX.W, dskinfo, 0);
	R->AF.B.h  = 0;	
	R->AF.B.l |= C_FLAG;
}



void fid_fd_read (int drive, Z80 *R)
{
	int err, n, seclen, head, cylinder;
	Z80 R1;

	if (!logged_in)
	{
		fid_fd_login(drive, &R1);
		if (R1.AF.B.h)
		{
			R->AF.B.h = R1.AF.B.h;
			R->BC.B.h = R1.BC.B.h;
			return;
		}
	}
	setup_our_param();
	seclen = sizefrom(dskinfo[4]);
	calc_track(R->HL.W, &cylinder, &head);
	err = bios_retry(2, drive, head, cylinder, 1 + R->DE.W, 1, secbuf);
	if (err)
	{
		R->AF.B.h = map_err(err);
		R->BC.B.h = err;
		R->AF.B.l &= ~C_FLAG;
		R->AF.B.l &= ~Z_FLAG;
		return;
	}
	for (n = 0; n < seclen; n++)
	{
		WrZ80(R->IY.W + n, secbuf[n]);
	}
	R->AF.B.h = 0;
	R->AF.B.l |= C_FLAG;
}


void fid_fd_write(int drive, Z80 *R)
{
	int err, n, seclen, head, cylinder;
	Z80 R1;

	if (!logged_in)
	{
		fid_fd_login(drive, &R1);
		if (R1.AF.B.h)
		{
			R->AF.B.h = R1.AF.B.h;
			R->AF.B.l = R1.AF.B.l;
			R->BC.B.h = R1.BC.B.h;
			return;
		}
	}

	setup_our_param();
	seclen = sizefrom(dskinfo[4]);
	for (n = 0; n < seclen; n++)
	{
		secbuf[n] = RdZ80(R->IY.W + n);
	}
	calc_track(R->HL.W, &cylinder, &head);
	err = bios_retry(2, drive, head, cylinder, 1 + R->DE.W, 1, secbuf);
	if (err)
	{
		R->AF.B.h = map_err(err);
		R->BC.B.h = err;
		R->AF.B.l &= ~C_FLAG;
		R->AF.B.l &= ~Z_FLAG;
		return;
	}
	R->AF.B.h = 0;
	R->AF.B.l |= C_FLAG;

}

void fid_fd_flush(int drive, Z80 *R) 
{
	R->AF.B.h = 0;
	R->AF.B.l |= C_FLAG;
}

char * fid_fd_mess (Z80 *R) 
{
	switch(R->BC.B.h)
	{
		case 1: return "Invalid parameter";
		case 2: return "Missing address mark";
		case 3: return "Disc read-only";
		case 4: return "Sector not found";
		case 5: return "Reset failed";
		case 6: return "Disc changed, please replace";
		case 8: return "DMA overrun";
		case 9: return "Data boundary error";
		case 12: return "Invalid media";
		case 16: return "CRC data error";
		case 32: return "Controller failed";
		case 64: return "Seek fail";
		case 128: return "Drive not ready";
	}
	return "Unknown error";
}


int  fid_fd_init(int drive)	/* Check if drive n exists */
{
	char str[20];
	FILE *fp;
	
	sprintf(str, "/dev/fd%d", drive);

	fp = fopen(str, "rb");
	if (!fp) return 0;

	fclose(fp); 
	return 1;	
}


void fid_fd_unit(void)
{
}


static void zputs(word W, char *s)
{
	int n = 0;

	while (n < 16 && (*s))
	{
		if (*s < ' ') *s = ' ';
		WrZ80(W++, *s);
		++s;
		++n;
	}
	while (n < 16)
	{
		WrZ80(W++, ' ');
		++n;
	}
}

/* Parse the file BOOT/BOOT.CFG */

void boot_id(Z80 *R)
{
	char line[50];
	char bootcfg[160];
	FILE *fp;

	strcpy(bootcfg, find_boot_file(BOOT_CFG));

	R->AF.B.h = 0;

	fp = fopen(bootcfg, "r");
	if (!fp)
	{
		sprintf(bootcfg,"boot%d.dsk", R->BC.B.h);
		strcpy(bootcfg, find_boot_file(bootcfg));
		fp = fopen(bootcfg,"rb");
		if (fp)
		{
			fclose(fp);
			sprintf(line,"boot%d.dsk", R->BC.B.h);
			zputs(R->HL.W, line);
			R->AF.B.h = 1;
		}
		else	R->AF.B.h = 0;
	}
	else	/* BOOT.CFG exists */
	{
		while (fgets(line, 50, fp))
		{
			char *s = strchr(line, '=');

			if (s && atoi(line) == R->BC.B.h)
			{
				zputs(R->HL.W, s + 1);
				R->AF.B.h = 1;
				break;
			}
		} 
		fclose(fp);
	}
}
