/** JOYCE PCW emulator for DJGPP *****************************/
/**                                                         **/
/** JOYCEFDC.C - uPD765A floppy disc controller             **/
/**                                                         **/
/** Copyright (c) John Elliott 1996-1998,2000               **/
/**     You are not allowed to sell this software for       **/
/**     profit.                                             **/   
/*************************************************************/

#include "Joyce.h"

byte fdctrace;

/* PC ports for the host FDC - currently unused */
#define FDC_SELECT 0x3F2
#define FDC_STATUS 0x3F4
#define FDC_DATA   0x3F5

/* Max no. of bytes that can be buffered by this simulated FDC */

#define MAX_SECTOR_LEN 8192

/* Cause of an FDC interrupt. 
 * 
 * 1 => Entering result phase of Read/Write/Format/Scan command
 * 2 => Ready for data transfer (in execution phase)
 * 4 => End of Seek or Recalibrate command
 */

static int int_cause;

/* The FDC status registers: st0, st1, st2, st3 */

static byte st0, st1, st2, st3;

/* The results from the SPECIFY command */

static byte spec1, spec2;

/* The last drive for which diskimage header info was requested */
/* Set this to -1 if any image file is closed */

static int last_info_request = -1;

/* The last sector for which a DD READ ID request was made. Reset by
  seeks, reads, writes etc. */

static int last_id_request = 0;

/* Buffers for data transfer in and out of the FDC */

static int cur_cmd = -1;	/* Command being read in, -1 for none */
static int cur_cnt = 0;		/* No. of bytes read so far in this command */
static int fdc_cmd[20];		/* Buffer for command bytes received */

static byte fdc_xfer[MAX_SECTOR_LEN];	/* Execution phase buffer */
static int fdc_xfercnt;			/* No. of bytes to transfer */
static int fdc_xfercur;			/* Bytes transferred so far */

static int fdc_rsl[20]; 		/* Result bytes to return */
static int fdc_rslcnt;			/* No. of bytes to transfer */
static int fdc_rslcur;			/* Bytes returned so far */

/* The lengths of the standard commands, 1 for unknown commands */

static int bytes_in_cmd[32] = { 1,  /* 0 = none */
				1,  /* 1 = none */  
				9,  /* 2 = READ TRACK, not implemented */
				3,  /* 3 = SPECIFY */
				2,  /* 4 = SENSE DRIVE STATUS */ 
				9,  /* 5 = WRITE DATA */ 
				9,  /* 6 = READ DATA */
				2,  /* 7 = RECALIBRATE */
				1,  /* 8 = SENSE INTERRUPT STATUS */ 
				9,  /* 9 = WRITE DELETED DATA, not implemented */
				2,  /* 10 = READ SECTOR ID */
				1,  /* 11 = none */
				9,  /* 12 = READ DELETED DATA, not implemented */
				6,  /* 13 = FORMAT A TRACK */
				1,  /* 14 = none */ 
				3,  /* 15 = SEEK */
				1,  /* 16 = none */
				9,  /* 17 = SCAN EQUAL */
				1,  /* 18 = none */
				1,  /* 19 = none */
				1,  /* 20 = none */
				1,  /* 21 = none */
				1,  /* 22 = none */
				1,  /* 23 = none */ 
				1,  /* 24 = none */
				9,  /* 25 = SCAN LOW OR EQUAL */
				1,  /* 26 = none */
				1,  /* 27 = none */
				1,  /* 28 = none */
				1,  /* 29 = none */
				9,  /* 30 = SCAN HIGH OR EQUAL */
				1   /* 31 = none */
			      };

static int more_bytes = 0;	/* No. of bytes remaining in current command */

static void fdc_write_end(void);	/* Called after data transferred in a Write command */
static void fdc_fmt_end(void);		/* Called after data transferred in a Format command */

static byte image_header[256];		/* The header of a CPCEMU-format disc image */
					/* Its magic number is "MV - CPC" at 0 */
					/* Some useful offsets in the header: */
#define OFFS_TRK  0x30	/* No. of tracks */
#define OFFS_HEAD 0x31	/* No. of heads */
#define OFFS_TLEN 0x32  /* Bytes / track, little-endian word */

#ifdef __MSDOS__
struct diskinfo_t bios_info;	/* parameters for a BIOS call */
#endif

/* Create a new, empty CPCEMU disc image */

int img_new(int unit)
{
	FILE *fp;
	int err;

	unit &= 1;
	fp = fopen(drive_filename[unit], "wb");
	if (!fp) 
	{
		joyce_dprintf("Cannot open %s\n", drive_filename[unit]);
		return 0;
	}

	memset(image_header, 0, 256);	
	sprintf((char *)image_header,"MV - CPCEMU Disk-File\r\nDisk-Info\r\n(JOYCE)");
	err = fwrite(image_header, 1 , 256, fp);
	fclose(fp);
	return (err == 256);
}

/* The following code is embryonic support for access to the host PC's
 * disc drive via the BIOS */


void bios_reset(void)
{
#ifdef __MSDOS__
	_bios_disk(_DISK_RESET, NULL);
#endif
}

byte bios_stat(void)	/* Get status of last operation */
{
#ifdef __MSDOS__
	int rv;
	rv = _bios_disk(_DISK_STATUS, &bios_info);
	return rv;
#else
	return 1;
#endif
}


int xlterr(byte error)	/* Translate a BIOS error */
{
	switch(error)
	{
		case 0: st0 = (st0 & 0x3F);			break;
		case 1: st0 = (st0 & 0x3F) | 0x80;		break;
		case 2: st0 = (st0 & 0x3F) | 0x40; st1 |= 1;	break;
		case 3: st0 = (st0 & 0x3F) | 0x40; st1 |= 2;	break;
		case 4: st0 = (st0 & 0x3F) | 0x40; st1 |= 4;	break;
		case 6: 
		case 0x80: st0 = (st0 & 0x3F) | 0xC8;		break;
		case 8:	st0 = (st0 & 0x3F) | 0x40; st1 |= 16;	break;
		case 16: st0 = (st0 & 0x3F)| 0x40; st2 |= 32;	break;
		case 0x40: st0 = (st0 & 0x3F) | 0x50;		break;

		default: st0 = (st0 & 0x3F);
	}
	if (!error) return 1;

	jstate.fdc_msr = 0xD0;	/* Ready to return results */
	fdc_rslcur = 0;
	fdc_rslcnt = 1;
	fdc_rsl[0] = st0;
	return 0;
}

byte bios_drvdata(byte drive, byte offs)	/* Get info for drive */
{
#ifdef __MSDOS__
	byte infobuf[4];

	biosdisk(8, (drive & 1), 0, 0, 0, 0, &infobuf);
	
	switch(offs)
	{
		case OFFS_TRK:  return infobuf[1];		// Cylinders
		case OFFS_HEAD: return infobuf[3];		// Heads
		case OFFS_TLEN: return (512 * infobuf[0]);	// Sectors
	}
#endif /* def __MSDOS__ */
	return 0;
}

void img_eject(int unit)	/* "eject" a disk */
{
	if (drive_fp[unit & 1]) fclose(drive_fp[unit & 1]);
#ifdef PCFLOPPY
	drive_fd[unit & 1] = -1;
#endif
	drive_fp[unit & 1] = NULL;
	drive_filename[unit & 1] = NULL;
	last_info_request = -1;
}

/* Is an image file "ready"? */
/* (this must be called before trying to do anything with it */

#define RET1 { st3 |= 0x20; return 1; }
#define RET0 { st3 &= 0xDF; return 0; }

int img_ready(int unit)
{
	if (!jstate.motor_on) RET0

#ifdef PCFLOPPY
	if (drive_fd[unit & 1] >= 0)
	{
		int st = bios_stat();
		switch(st)
		{
			case 0x80: case 0x06: RET0
			default: RET1
		}
	}	
	else
#endif
	{
		if (drive_fp[unit & 1]) RET1	  /* Image already open*/

		if (!drive_filename[unit & 1]) RET0   /* Image has no filename */
		/* Try to open it */	
		drive_fp[unit & 1] = fopen(drive_filename[unit & 1], "r+b");
        	if (!drive_fp[unit & 1]) 
       		{
	                joyce_dprintf("Cannot open %s read-write - trying read-only\n"
				, drive_filename[unit & 1]);

			drive_fp[unit & 1] = fopen(drive_filename[unit & 1], 
                                                   "rb");
			if (drive_fp[unit & 1]) drive_readonly[unit & 1] = 1;
	      		else joyce_dprintf("Can't even open read-only\n"); 
		 }
		if (drive_fp[unit & 1]) RET1 /* Open succeeded */
	}
	RET0
}

#undef RET1
#undef RET0

/* Retrieve a byte from the MV - CPC header. */

int img_info(int drive, int offs)
{
#ifdef PCFLOPPY
	if (drive_fd[drive & 1] >= 0)
	{
		return bios_drvdata(drive & 1, offs);		
	}
#endif
	if (!img_ready(drive & 1)) return -1;

	if ((drive & 1) != last_info_request)
	{
		fseek(drive_fp[drive & 1], 0, SEEK_SET);
		fread(image_header, 1, 256, drive_fp[drive & 1]);
		last_info_request = drive & 1;
	}
	if (memcmp("MV - CPC", image_header, 8)) return -1; /* Not an MV-CPC image */

	return image_header[offs];
}

#define MAX_CYL(drive)  img_info(drive, OFFS_TRK)
#define MAX_HEAD(drive) img_info(drive, OFFS_HEAD)
#define TRKLEN(drive)   img_info(drive, OFFS_TLEN) + (256 * img_info(drive, OFFS_TLEN + 1))

/* Write a byte to the MV-CPC header */

static void set_imgdata(byte drive, byte offs, byte value)
{
	if (!img_ready(drive & 1)) return;

	if ((drive & 1) != last_info_request)
	{
		fseek(drive_fp[drive & 1], 0, SEEK_SET);
		fread(image_header, 1, 256, drive_fp[drive & 1]);
		last_info_request = drive & 1;
	}
	if (memcmp("MV - CPC", image_header, 8)) return; /* Not an MV-CPC image */

	image_header[offs] = value;

	fseek(drive_fp[drive & 1], 0, SEEK_SET);
	fwrite(image_header, 1, 256, drive_fp[drive & 1]);
}


#define SET_TRKLEN(dr, tl) { set_imgdata(dr, OFFS_TLEN,     tl & 0xFF); \
			     set_imgdata(dr, OFFS_TLEN + 1, tl >> 8); }
#define SET_HCOUNT(dr, hc) set_imgdata(dr,OFFS_HEAD,hc)
#define SET_MAXTRK(dr, tc) set_imgdata(dr,OFFS_TRK, tc)



static int last_d, last_h;

long img_seek(int drive, int cylinder, int head)	/* Seek to a track in */
{							/* a CPCEMU image */
	long offs; 
	int hcount = MAX_HEAD(drive);

	last_d = drive;
	last_h = head;

	if (head >= hcount) return -1;
	
	offs = (cylinder * hcount) + head;	/* Where the track is in the
						 * image */
	offs *= TRKLEN(drive);			/* Length of track */
	offs += 0x100;				/* Length of header */

	fseek(drive_fp[drive & 1], offs, SEEK_SET);

	return offs;
}

/* << v1.36 */
long img_reseek(int cylinder)
{
	return img_seek(last_d, cylinder, last_h);
}

/* >> v1.36 */

static byte trackinfo[256];

void mk_trkinfo(void)	/* Create a blank CPCEMU track-info block */
{
	memset(trackinfo, 0, 256);
	sprintf((char *)trackinfo,"Track-Info\r\n");
}


long img_vrfy(void)	/* Verify a CPCEMU track-info block */
{
	int n, done;
	long seclen;
	byte *secbuf;

	do
	{
		done = 1;

		fread(trackinfo, 1, 256, drive_fp[jstate.fdc_unit & 1]);
	
		if (memcmp(trackinfo,"Track-Info", 10))
		{
			joyce_dprintf("BAD DSK file: missing track %d "
				      "header in %s\n", fdc_cmd[2],
                               drive_filename[jstate.fdc_unit & 1]);

			/* Missing address mark */
			st2 |= 1;
			st1 |= 0;
			st0 |= 0x40;
			return -1;
		}

		secbuf = trackinfo + 0x18;
		for (n = 0; n < trackinfo[0x15]; n++)
		{
			if (secbuf[2] == fdc_cmd[4]) break; 
					/* Found our sector */ 
			secbuf += 8;		
		}
		if (n == trackinfo[0x15])
		{
			joyce_dprintf("Sector %d not found. \n", fdc_cmd[4]);
			st0 |= 0x40;	/* Error code */
			st1 |= 0x82;	/* End of cylinder, and 
					 * No Data [sector ID not found] */
			return -1;
		}

/* << v1.36: If heads match but track found is 2 * track requested, then
 *          CP/M is likely trying to double step the drive. Since DSK 
 *          images don't distinguish between single and double track 
 *          discs, we previously assumed they always had the same 
 *          "trackness" as the drive. This code would allow emulation of 
 *          double stepping - ST discs in a DT drive. However, it only works
 *          for the first half of the disc (not the second) because 
 *          there are no track headers for tracks off the end of the disc.
 *          And with no track headers, we can't tell if we've gone twice
 *          as far as we should.

		if (secbuf[1] == fdc_cmd[3] &&
		    secbuf[0] != 0 &&
		    secbuf[0] == (2 * fdc_cmd[2]))
		{
			// Move the track to under the disc head!
			long offset = img_reseek(fdc_cmd[2]);		

			if (offset < 0)
			{
                		st2 |= 2;
				st0 |= 0x40;
				return -1;
			}
			joyce_dprintf("Attempting to compensate for double "
				      "stepping on track %d -> %d\n", 
                                      secbuf[0], fdc_cmd[2]);
			done = 0;	// Restart this 
		} */
	}
	while (!done);	 
/* >> v1.36 */

	if ((secbuf[0] != fdc_cmd[2])   /* Wrong cylinder */
	||  (secbuf[1] != fdc_cmd[3]))  /* or wrong head */
	{
		joyce_dprintf("Looking for cylinder %d head %d - found "
                              " %d,%d\n", fdc_cmd[2], fdc_cmd[3],
                              secbuf[0], secbuf[1]);
		st2 |= 0x10;
		st0 |= 0x40;
		return -1;
	}

	if (secbuf[3] != fdc_cmd[5]) /* Sector size wrong */
	{
		st0 |= 0x40;
		st1 |= 4;	/* No data */
		return -1;
	}	

	seclen = (128 << fdc_cmd[5]); /* Size of sector, bytes */

	/* We've found the sector */

	fseek(drive_fp[jstate.fdc_unit & 1], n * seclen ,SEEK_CUR);
	return n * seclen;
}


void img_savesec(void)	/* Write a sector to a CPCEMU image */
{
	int seclen = (128 << fdc_cmd[5]); /* Size of sector, bytes */


	/* Deal with multiple sectors 
	if (fdc_cmd[4] != fdc_cmd[6]) seclen *= abs(fdc_cmd[6] - fdc_cmd[4]);
	*/

	if (fdc_cmd[8] < 0xFF) seclen = fdc_cmd[8];	

	if (fwrite(fdc_xfer, 1, seclen, drive_fp[jstate.fdc_unit & 1]) < seclen)
	{
		st1 |= 2;
	}
}


void img_loadsec(void)	/* Read a sector from a CPCEMU image */
{
	int seclen = (128 << fdc_cmd[5]); /* Size of sector, bytes */

	/* Deal with multiple sectors 
	if (fdc_cmd[4] != fdc_cmd[6]) seclen *= abs(fdc_cmd[6] - fdc_cmd[4]);
	*/

	if (fdc_cmd[8] < 0xFF) seclen = fdc_cmd[8];	

	if (fread(fdc_xfer, 1, seclen, drive_fp[jstate.fdc_unit & 1]) < seclen) 
	{
		st1 |= 0x20;	/* Data error */
		st2 |= 0x20;
	}

	fdc_xfercur = 0;
	fdc_xfercnt = seclen;
}


static void fdc_res7(void)	/* Return 7 result bytes */
{
	fdc_rsl[0] = st0;
	fdc_rsl[1] = st1;
	fdc_rsl[2] = st2;
	fdc_rsl[3] = fdc_cmd[2];	/* C, H, R, N */
	fdc_rsl[4] = fdc_cmd[3];
	fdc_rsl[5] = fdc_cmd[4];
	fdc_rsl[6] = fdc_cmd[5];
	jstate.fdc_msr = 0xD0;		/* Ready to return results */
}

static void fdc_resint(void)	/* Start of result phase, interrupt */
{
	jstate.fdcint = 1000;	/* (v1.30) Make the FDC slower, or it can */
				/* overtake the CPU and cause a crash */
	int_cause = 1;		/* Int caused by result phase */	
}


static void fdc_execint(void)	/* Execution phase interrupt */
{
	jstate.fdcint = 1000;	/* (v1.30) Make the FDC slower */
	int_cause = 2; 		/* Int caused by execution phase */
}	


static void scan_cmp(byte fdcbyte, byte cpubyte)	/* Compare two bytes */
{							/* for the SCAN cmd */
	int cmd = fdc_cmd[0] & 0x1F;
	if ((st2 & 0x0C) == 8) /* Scan equal is still set */	
	{
		if ((fdcbyte == cpubyte) || 
		    (fdcbyte == 0xFF)    || 
		    (cpubyte == 0xFF)) return;
		
		/* They differ */
		if (cmd == 17) /* equal */
		{
			st2 = (st2 & 0xF3) | 4;	/* Unequal */
			return;
		}
		if (cmd == 25) /* low or equal */
		{
			if (fdcbyte < cpubyte) st2 = (st2 & 0xF3);
			if (fdcbyte > cpubyte) st2 = (st2 & 0xF3) | 4;
			st2 = (st2 & 0xF3);
			return;
		}
		if (cmd == 30) /* high or equal */
		{
			if (fdcbyte < cpubyte) st2 = (st2 & 0xF3) | 4;
			if (fdcbyte > cpubyte) st2 = (st2 & 0xF3);
			st2 = (st2 & 0xF3);
			return;
		}
	}
}

static void fdc_endexec(void)
{
	byte theCmd = fdc_cmd[0] & 0x1F;

	jstate.fdc_msr = 0xD0;	/* End execution phase, start to return results */
	fdc_rslcur = 0;

	switch(theCmd)
	{
		case 17:			  /* SCAN */
		case 25:
		case 30:  fdc_res7();             /* FALL THROUGH */
		case 6:   fdc_rslcnt = 7; break;  /* READ */

		case 5:   fdc_write_end();
			  fdc_rslcnt = 7; break;  /* WRITE */

		case 13:  fdc_fmt_end();
			  fdc_rslcnt = 7; break;  /* FORMAT */
	}
}

static void fdc_error(void)
{
	st0 = (st0 & 0x3F) | 0x80; /* Invalid command */
	jstate.fdc_msr = 0xD0;	/* Ready to return results */
	fdc_rslcur = 0;
	fdc_rslcnt = 1;
	fdc_rsl[0] = st0;
}


static void fdc_endcmd(void)
{
	jstate.fdc_msr = 0x80; /* End command, return to standby */
	if (int_cause < 3) int_cause = 0;
	cur_cmd = -1;
}

static void fdc_getdrive(void)
{
	st0 = (st0 & 0xF8) | (fdc_cmd[1]&0x7);
	st3 = (st3 & 0xF8) | (st0 & 7);
	jstate.fdc_unit = fdc_cmd[1]&3;
	jstate.fdc_head = (fdc_cmd[1]&4) >> 2;
}


/* --------- FDC command handlers ----------- */

static void fdc_format(void)
{
	long offs;
	int seclen = (128 << fdc_cmd[2]); /* Size of sector, bytes */
	int hcount;
	int unit;

	last_id_request = 0;

	st0 = st1 = st2 = 0;
	fdc_getdrive();
	
	unit = jstate.fdc_unit & 1;

	hcount = MAX_HEAD(unit);
	if (jstate.fdc_head >= hcount) 
	{
		hcount = jstate.fdc_head + 1;
		SET_HCOUNT(unit, hcount);
	}
	SET_TRKLEN(unit, (0x100 + (seclen * fdc_cmd[3])));
	if (MAX_CYL(unit) <= drive_cylinder[unit]) SET_MAXTRK(unit, drive_cylinder[unit] + 1);		

	offs = (drive_cylinder[unit] * hcount) + jstate.fdc_head;	
						 /* Where the track is in the
						  * image */
	offs *= ((seclen * fdc_cmd[3]) + 0x100); /* Length of track */
	offs += 0x100;				 /* Length of header */

	fseek(drive_fp[unit], offs, SEEK_SET);

	/* This has done an fseek() to the start of the track */

	mk_trkinfo();
	
	if (drive_readonly[unit]) 	/* R/O, can't write */
	{
		st1 |= 2;	/* Read only */
		st0 |= 0x40;
		fdc_endexec();
		fdc_res7();
		fdc_resint();
		return;
	}

	fdc_execint();	
	jstate.fdc_msr	= 0xB0;	/* FDC is ready to transfer data... */
	fdc_xfercur = 0;
	fdc_xfercnt = 4*fdc_cmd[3];
}

static void fdc_fmt_end(void)
{
	int n;
	int seclen = (128 << fdc_cmd[2]); /* Size of sector, bytes */

	trackinfo[0x10] = fdc_xfer[0];
	trackinfo[0x11] = fdc_xfer[1];
	trackinfo[0x14] = fdc_cmd[2];
	trackinfo[0x15] = fdc_cmd[3];
	trackinfo[0x16] = fdc_cmd[4];
	trackinfo[0x17] = fdc_cmd[5];

	for (n = 0; n < fdc_cmd[3]; n++)
	{
		trackinfo[0x18 + 8*n] = fdc_xfer[0x00 + 4*n];
		trackinfo[0x19 + 8*n] = fdc_xfer[0x01 + 4*n];
		trackinfo[0x1A + 8*n] = fdc_xfer[0x02 + 4*n];
		trackinfo[0x1B + 8*n] = fdc_xfer[0x03 + 4*n];
	}
	if (fwrite(trackinfo, 1, 256, drive_fp[jstate.fdc_unit & 1]) < seclen) st1 |= 2;

	memset(fdc_xfer, fdc_cmd[5], seclen);

	for (n = 0; n < fdc_cmd[3]; n++)
	{
		if (fwrite(fdc_xfer, 1, seclen, drive_fp[jstate.fdc_unit & 1]) < seclen)  st1 |= 2;	
	}
	fdc_res7();
	fdc_resint();
}

static void fdc_read(void)
{
	long offs;

	last_id_request = 0;

	st0 = st1 = st2 = 0;
	fdc_getdrive();

#ifdef PCFLOPPY
	if (drive_fd[jstate.fdc_unit & 1] >= 0)
	{
		if (xlterr(biosdisk(2, drive_fd[jstate.fdc_unit & 1], 
			fdc_cmd[3],	/* Head */
			fdc_cmd[2],	/* track */
			fdc_cmd[4],	/* sector */
			1,		/* 1 sector */
			fdc_xfer)))			

		{
			fdc_res7();
			fdc_execint();	
			jstate.fdc_msr	= 0xF0;	/* FDC is ready to transfer data... */
			fdctrace = 1;
		}
		else
		{
			fdc_endexec();
			fdc_res7();
			fdc_resint();
		}
		return;
	}
#endif /* PCFLOPPY */
	offs = img_seek(jstate.fdc_unit & 1, drive_cylinder[jstate.fdc_unit & 1], jstate.fdc_head);	

	if (offs < 0)
	{
		st2 |= 2;
		st0 |= 0x40;
		fdc_endexec();
		fdc_res7();
		fdc_resint();
		return;
	}
	/* This has done an fseek() to the start of the track */
	offs = img_vrfy();

	if (offs < 0) /* Can't find the cylinder. This sets st0-3 */
	{
		fdc_endexec();
		fdc_res7();
		fdc_resint();
		return;
	}
	img_loadsec();
	
	fdc_res7();
	fdc_execint();	
	jstate.fdc_msr	= 0xF0;	/* FDC is ready to transfer data... */
}



static void fdc_scan(void)	/* Set up for scan type command */
{
	long offs;
	int seclen = (128 << fdc_cmd[5]); /* Size of sector, bytes */

	last_id_request = 0;

	st0 = st1 = st2 = 0;
	fdc_getdrive();

	offs = img_seek(jstate.fdc_unit & 1, drive_cylinder[jstate.fdc_unit & 1], jstate.fdc_head);	

	if (offs < 0)
	{
		st2 |= 2;
		st0 |= 0x40;
		fdc_endexec();
		fdc_res7();
		fdc_resint();
		return;
	}
	/* This has done an fseek() to the start of the track */

	offs = img_vrfy();

	if (offs < 0) /* Can't find the cylinder. This sets st0-3 */
	{
		fdc_endexec();
		fdc_res7();
		fdc_resint();
		return;
	}
	img_loadsec();

	fdc_res7();
	fdc_execint();	
	st2 |= 0x08;	/* Scan Equal Hit */
	jstate.fdc_msr	= 0xB0;	/* FDC is ready to receive data... */
	fdc_xfercur = 0;
	fdc_xfercnt = seclen;	/* Always a full sector */
}


static void fdc_write(void)
{
	long offs;
	int seclen = (128 << fdc_cmd[5]); /* Size of sector, bytes */

	last_id_request = 0;

	st0 = st1 = st2 = 0;
	fdc_getdrive();

	offs = img_seek(jstate.fdc_unit & 1, drive_cylinder[jstate.fdc_unit & 1], jstate.fdc_head);	

	if (offs < 0)
	{
		st2 |= 2;
		st0 |= 0x40;
		fdc_endexec();
		fdc_res7();
		fdc_resint();
		return;
	}
	/* This has done an fseek() to the start of the track */

	offs = img_vrfy();

	if (offs < 0) /* Can't find the cylinder. This sets st0-3 */
	{
		fdc_endexec();
		fdc_res7();
		fdc_resint();
		return;
	}
	
	if (drive_readonly[jstate.fdc_unit & 1]) 	/* R/O, can't write */
	{
		st1 |= 2;	/* Read only */
		st0 |= 0x40;
		fdc_endexec();
		fdc_res7();
		fdc_resint();
		return;
	}

	fdc_execint();	
	jstate.fdc_msr	= 0xB0;	/* FDC is ready to transfer data... */
	fdc_xfercur = 0;
	fdc_xfercnt = seclen;
}


static void fdc_write_end(void)
{
	img_savesec();
	
	fdc_res7();
	fdc_resint();
}

static void fdc_read_id(void)
{
	long offs;
	int n;

	fdc_rslcnt = 7;
	st0 = st1 = st2 = 0;
	fdc_getdrive();
	offs = img_seek(jstate.fdc_unit & 1, drive_cylinder[jstate.fdc_unit & 1], jstate.fdc_head);	
	if (offs < 0)
	{
		st1 |= 1;
		st0 |= 0x40;
		fdc_endexec();
		fdc_res7();
		fdc_resint();
		return;
	}
	/* This has done an fseek() to the start of the track */

	fread(trackinfo, 1, 256, drive_fp[jstate.fdc_unit & 1]);

	if (memcmp(trackinfo,"Track-Info", 10))
	{
                joyce_dprintf("BAD DSK file: missing track %d header in %s\n",
                               fdc_cmd[2],
                               drive_filename[jstate.fdc_unit & 1]);
	
		/* Missing address mark */
		st2 |= 1;
		st1 |= 0;
		st0 |= 0x40;
		return;
	}

	offs = 0x18 + 8*(last_id_request % trackinfo[0x15]);
	++last_id_request;

	/* << v1.36: memcpy() from byte[] to int[] doesn't quite do what
                    we want! 

	memcpy(fdc_cmd+2, trackinfo+offs, 4);	*/

	for (n = 0; n < 4; n++) fdc_cmd[2+n]  = trackinfo[offs+n];
	/* >> v1.36 */
	fdc_res7();
	fdc_resint();	
	fdc_endexec();
}



static void fdc_sis(void)
{
	if (int_cause > 2) /* FDC interrupted, and is ready to return status */
	{
		fdc_rsl[0] = st0;
		fdc_rsl[1] = drive_cylinder[st0 & 1];
		fdc_rslcnt = 2;	/* 2 bytes */
		fdc_endexec();
		jstate.inf8 &= 0xDF;	/* Clear the interrupt line */
		jstate.fdcint = 0;
		int_cause = 0;
	}
	else	/* FDC did not interrupt, error */
	{
		st0 = fdc_rsl[0] = 0x80;
		fdc_rslcnt = 1;
		fdc_endexec();
		jstate.inf8 &= 0xDF;	/* Clear the interrupt line */
		jstate.fdcint = 0;
		int_cause = 0;
	}
}



static void fdc_sds(void)
{
	fdc_getdrive();
	if (drive_readonly[jstate.fdc_unit & 1]) 
		st3 |= 0x40;
	else	st3 &= 0xBF;

	if (drive_cylinder[jstate.fdc_unit & 1] == 0) 
		st3 |= 0x10;
	else	st3 &= 0xEF;


	if (MAX_HEAD(jstate.fdc_unit & 1) > 1)
		st3 |= 8;
	else	st3 &= 0xF7;

	fdc_rsl[0] = st3;
	fdc_rslcnt = 1;
	fdc_endexec();
}

static void fdc_spec(void)
{
	spec1 = fdc_cmd[1];
	spec2 = fdc_cmd[2];
	fdc_endcmd();
}

static void fdc_seek(void)
{
	int cylinder = fdc_cmd[2];
	
	st0 = st1 = st2 = 0;

	fdc_getdrive();

	last_id_request = 0;

	drive_cylinder[jstate.fdc_unit & 1] = cylinder;
	fdc_endcmd();
	jstate.fdcint = 1000;	/* (v1.30) Make the FDC slower */
	int_cause = 4;	/* End of seek */
	st2 &= 0xFD;
	st0 |= 0x20;
	if (!img_ready(jstate.fdc_unit)) st0 |= 0x48;

	if (cylinder > MAX_CYL(jstate.fdc_unit)) 
	{
		st2 |= 2;
		st0 |= 0x40;
	}
}


static void fdc_recalibrate(void)
{
	st0 = st1 = st2 = 0;

	fdc_getdrive();

	last_id_request = 0;

	drive_cylinder[jstate.fdc_unit & 1] = 0;
	fdc_endcmd();
	jstate.fdcint = 1000;		/* (v1.30) Make the FDC slower */
	int_cause = 4;  /* End of seek */
	st2 &= 0xFD;
	st0 |= 0x20;
	if (!img_ready(jstate.fdc_unit))
	{
		st0 |= 0x48;	/* Not ready */
	}
}


/* --- FDC: main command dispatcher --- */

static void fdc_exec(void)	/* Start the execution phase */
{
/* This code to dump out FDC commands as they are received is very useful
 * in debugging
 *
 *
 *	int NC;
 *	joyce_dprintf("FDC: ");
 * 	for (NC = 0; NC < 9; NC++) joyce_dprintf("%02x ", fdc_cmd[NC]);
 *	joyce_dprintf("\n");
 */

	st0 &= 0xBF;	/* Reset "Seek finished" bit */
	switch(fdc_cmd[0] & 0x1F) /* Command */
	{
		/* Specify */
		case 3:  fdc_spec();   break;

		/* Sense drive status */
		case 4:  fdc_sds();    break;

		/* Write data */
		case 5:  fdc_write();   break;

		/* Read data */
		case 6:  fdc_read();   break;

		/* Recalibrate */
		case 7:  fdc_recalibrate(); break;

		/* Sense interrupt status */
		case 8:  fdc_sis();    break;

		/* Read ID */
		case 10: fdc_read_id(); break;
		
		/* Format a track */
		case 13: fdc_format(); break;
		
		/* Seek */
		case 15: fdc_seek();   break;

		/* Scan */
		case 17:
		case 25:
		case 30: fdc_scan();   break;

		/* Unknown */
		default: joyce_dprintf("Unknown FDC command %d\n", fdc_cmd[0] & 0x1F);
			 fdc_error();  break;
	}

}



void InitFDC(int type)    /* Initialise FDC emulation */
{
	int n;
	for (n = 0; n < 4; n++)
	{
		drive_cylinder[n] = 0;
		drive_fp[n] = NULL;
		if (drive_filename[n])  drive_readonly[n] = 0;
		else		 	drive_readonly[n] = 1;

#ifdef PCFLOPPY
		drive_fd[n] = -1;
#endif
	}
	if (jset.one_drive) drive_readonly[1] = 0;
	jstate.motor_on = 1;
	jstate.fdc_msr=0x80;
	st0 = 0;	/* Last command completed OK */
	st1 = 0;	/* Disk is R/W */
	st2 = 0;
	st3 = 0x30;	/* Single sided, ready, track 0 */
	jstate.inf8 |= 0x20;
}


void UnitFDC(void)    /* Deinit FDC emulation */
{
#ifdef __MSDOS__
	union REGS regs;

	regs.x.ax=0;
	regs.x.dx=0;
	int86(0x13,&regs,&regs);	/* Reset from BIOS */
#endif
}


/* Deal with an OUT to the FDC data register */
/* - reads in a command string */

void FDCport(unsigned char port, unsigned char value)
{

	if (!(port & 1)) return;	/* Can't write to status reg */
	if (int_cause > 0 && int_cause < 3)
	{
		jstate.fdcint = 0;
		jstate.inf8 &= 0xDF;
		int_cause = 0;
	}
	if (jstate.fdc_msr & 0x20) /* Execution phase */
	{
		if ((fdc_cmd[0] & 0x1F) == 17 ||
		    (fdc_cmd[0] & 0x1F) == 25 ||
		    (fdc_cmd[0] & 0x1F) == 30)	/* Scan type command */
		{
			scan_cmp(fdc_xfer[fdc_xfercur++], value);
		}
		else fdc_xfer[fdc_xfercur++] = value;
		fdc_xfercnt--;
		if (fdc_xfercnt == 0) 
		{
			fdc_endexec();
			fdc_resint();
		}
		if (int_cause > 0 && int_cause < 3)
		{
			jstate.fdcint = 1000;	/* (v1.30) Make the FDC slower */
		}
	}
	else if (cur_cmd < 1)
	{
		cur_cmd = value;
		cur_cnt = 1;
		fdc_cmd[0] = cur_cmd;
		more_bytes = (bytes_in_cmd[cur_cmd & 0x1F] - 1);
		if (more_bytes == 0) fdc_exec();
		else if (more_bytes < 0) fdc_error();
		jstate.fdc_msr |= 0x10;	/* In a command */
		return;
	}
	else
	{
		fdc_cmd[cur_cnt] = value;
		++cur_cnt;
		--more_bytes;
		if (more_bytes == 0) fdc_exec();	/* Put FDC into the
							 * execution phase */		
	}
}



byte FDCin(unsigned char port)
{
	byte fret = jstate.fdc_msr;

	if (port & 1)
	{
		if (int_cause > 0 && int_cause < 3)
		{
			jstate.fdcint = 0;
			jstate.inf8 &= 0xDF;
		}
		if (jstate.fdc_msr & 0x40) /* FDC ready to give data */
		{
			byte v;

			if (jstate.fdc_msr & 0x20) /* Execution phase */
			{
				v = fdc_xfer[fdc_xfercur++];
				fdc_xfercnt--;
				if (fdc_xfercnt == 0) 
				{
					fdc_endexec();
					fdc_resint();
				}
				if (int_cause > 0 && int_cause < 3)
				{
					jstate.fdcint = 1000;	/* (v1.30) Make the FDC slower */
				}
				return v;
			}
			else
			{
				v = fdc_rsl[fdc_rslcur++];
				fdc_rslcnt--;
				if (fdc_rslcnt == 0) fdc_endcmd();
				return v;						
			}
		}
	}
	else return jstate.fdc_msr;
	switch (st0 & 3)
	{
		case 0: fret = jstate.fdc_msr | 0x11;
		case 1: fret = jstate.fdc_msr | 0x12;
		case 2: fret = jstate.fdc_msr | 0x14;
		case 3: fret = jstate.fdc_msr | 0x18;
	}
	return fret;
}



void fdc_mstart(void)
{
	if (!jstate.motor_on)
	{
		jstate.fdcint = 1333333L; /* FDC interrupts after 1 sec */
		st0 &= 0xF7;	/* Remove "not ready" from ST0 */
		st3 |= 0x20;	/* Set "ready" on ST3 */
	}
	jstate.motor_on = 1;
}

void fdc_mstop(void)
{
	if (jstate.motor_on)
	{
		jstate.fdcint = 1333333L; /* FDC interrupts after 1 sec */
		st0 |= 0x08;	/* Set "not ready" */
		st3 &= 0xDF;	/* Clear "ready" on ST3 */
		if ((jstate.fdc_msr & 0xF0) != 0x80)	/* FDC is doing something */
		{
			st0 |= 0xC0;	/* Command failed, drive stopped */
			fdc_endexec();
		}
		memset(drive_cylinder, 0, sizeof(drive_cylinder));
	}
	jstate.motor_on = 0;
}

void fdc_set_tcount(unsigned char tc)
{
	if (tc == 5 && (jstate.fdc_msr & 0x20))	/* In execution phase? */
	{
		fdc_endexec();
	}
}



int fdc_boot(int image)
{
	static char filename[160];
	byte cksum;
	int n;
	long seclen;
	byte *secbuf;

	sprintf(filename,"boot%d.dsk", image);
	strcpy(filename, find_boot_file(filename));

	joyce_dprintf("Loading %s\n", filename);
	
	img_eject(0);
	drive_readonly[0] = jset.boot_ro;
	drive_filename[0] = filename;		

	if (jset.one_drive)	/* Disable B: if told to */
	{
		img_eject(1);
		drive_readonly[1] = 0;
	}

	/* Load the boot sector */

	fdc_mstart();
	if (!img_ready(0)) return 0;

	joyce_dprintf("Image file exists.\n");

/* DEBUGGING CODE */	//drive_fd[0] = 0;

#ifdef PCFLOPPY	
	if (drive_fd[0] >= 0)
	{
		int n;
		for (n = 0; n < 3; n++)
		{
			int nerr;
			nerr = biosdisk(2, drive_fd[0], 0, 0, 1, 1, PCWRAM + 0xF000);
			if (!nerr) break;
			biosdisk(0, drive_fd[0], 0, 0, 0, 0, NULL);
		}
	}
	else
#endif
	{
		fseek(drive_fp[0], 256, SEEK_SET);	/* Track 0 info */

		fread(trackinfo, 1, 256, drive_fp[0]);
	
		if (memcmp(trackinfo,"Track-Info", 10)) return 0;

		secbuf = trackinfo + 0x18;
		for (n = 0; n < trackinfo[0x15]; n++)
		{
			if (secbuf[2] == 1) break; /* Found our sector */ 
			secbuf += 8;		
		}
		if (n == trackinfo[0x15]) return 0;

		if (secbuf[0] || secbuf[1]) return 0; /* Wrong track/head */

		if (secbuf[3] != 2) return 0; /* Sector size wrong */

		seclen = 512; /* Size of sector, bytes */

	/* We've found the sector */

		fseek(drive_fp[0], n * seclen, SEEK_CUR);
		fread(PCWRAM + 0xF000, 512, 1, drive_fp[0]);
	}
	cksum = 0;
	for (n = 0xF000; n < 0xF200; n++)
	{
		cksum += PCWRAM[n];
	}	
	jstate.model_type = 0;
	if (cksum == 1)		/* v1.20 : Can boot 9512 discs */
	{
		joyce_dprintf("Warning: 9000 boot disc\n");
		jstate.model_type = 1; 
	}
	else if (cksum != 0xFF) 
	{
		popup_on("Not an 8000 boot disc", 4, 1); /* (v1.27) */
		return 0;
	}
	memset(PCWRAM, 0, 0xF000);

	return 1;
}
	

