/** JOYCE PCW emulator for DJGPP *****************************/
/**                                                         **/
/** JOYCEFID: Support for JOYCEDRV.FID and VGA.FID, which   **/
/**          are Z80 stubs to the C functions in JOYCE.     **/
/**                                                         **/
/** Copyright (c) John Elliott 1996-1998                    **/
/**     You are not allowed to sell this software for       **/
/**     profit.                                             **/   
/*************************************************************/

#include "Joyce.h"
#include "Joycefid.h"
#include "Joycevga.h"
#include "Joycedsk.h"	/* v1.22 */
#include "inline.h"

/* DPB for an 8Mb hard drive */
static const byte dpb[] = { 0x80, 0x00, 0x05, 0x1F, 0x01, 0xFF, 0x07, 0xFF,
			    0x03, 0xFF, 0x00, 0x00, 0x80, 0x00, 0x00, 0x02,
			    0x03 };

char dskfil[] = "x.dsk";
char dskfilu[] = "X.DSK";
char hooked[49];
char namebuf[161];
byte allow_floppy = 0;
char drives[16] = { -1, -1, -1, -1, -1, -1, -1, -1,
		    -1, -1, -1, -1, -1, -1, -1, -1 };
			/* v1.22   drives[x] is 0-3 for a floppy, else -1 */

static FILE *images[16];
static word fidtext;
static byte lastdrv;

static byte isdisk(char *s)	/* v1.22 special file for floppy? */
{
	if (!strncasecmp(s,"DISC:", 5)) return 1;
	if (!strncasecmp(s,"DISK:", 5)) return 1;
	return 0;
}

static void fid_puts(word addr, char *s)
{
	while (*s) WrZ80(addr++, *s++);
	WrZ80(addr++, 0x0D);
	WrZ80(addr++, 0x0A);
	WrZ80(addr++, 0xFF);
}

static void fid_puts1(word addr, char *s)
{
	while (*s) WrZ80(addr++, *s++);
	WrZ80(addr++, 0xFF);
}



static int open_image(int img)
{
	char *dname;
	
	dskfil[0]  = img + 'a';
	dskfilu[0] = img + 'A'; 
	if (drives[img] >= 0) return 1;
	if (images[img])      return 1;
	dname = getenv(dskfil);
/* v1.34: Do case-insensitive check on env. variables */
	if (!dname) dname = getenv(dskfilu);
	if (!dname)
	{
		strcpy(namebuf, find_dsk_file(dskfil));
		dname = namebuf;
	}
	drives[img] = -1;			/* Not a BIOS drive */
	if (isdisk(dname))
						/* v1.22: special file for */
	{					/* BIOS floppy access */
		int unit = atoi(dname + 5);

		if (!allow_floppy) return 0;
		if (!fid_fd_init(unit)) return 0;
		drives[img] = unit;
		return 1;
	}
	if ((images[img] = fopen(dname, "r+b")))
	{
		int nb;
		for (nb = 0; nb < 256; nb++)
		if (fgetc(images[img]) != 0xE5)
		{
			fclose(images[img]);
			images[img] = NULL;
			return 0;
		}
		return 1;
	}
	if ((images[img] = fopen(dname, "w+b")))
	{
		int nb;
		for (nb = 0; nb < 256; nb++) 
		{
			if (fputc(0xE5, images[img]) == EOF)
			{
				fclose(images[img]);
				images[img] = NULL;
				return 0;
			}
		}
	}
	return 1;
}

static int close_image(int img)
{
	if (!images[img]) return 1;
	fclose(images[img]);
	images[img] = NULL;
	return 1;
}


static void ret_err(Z80 *R)
{
	R->AF.B.h = 1;
	R->AF.B.l &= ~C_FLAG;
	R->AF.B.l |= Z_FLAG;
	R->AF.B.l &= ~H_FLAG;
}

static void ret_ok(Z80 *R)
{
	R->AF.B.h = 0;
	R->AF.B.l |= C_FLAG;
	R->AF.B.l &= ~H_FLAG;
}


void fid_cleanup(void)
{
	int b;
	for (b=0; b<16; b++) close_image(b);
}

void fid_d_login(Z80 *R)
{
	int nx;
	byte drv = R->BC.B.h;	/* (v1.22) check for floppy access */
	
	lastdrv = drv;
	if (!open_image(drv)) 
	{
		ret_err(R);
		return;
	}
	if (drives[drv] >= 0) 
	{
		for (nx = 0; nx < 17; nx++) WrZ80(R->IX.W + nx, 0);
		fid_fd_login(drives[drv], R);
		return;
	}
	for (nx = 0; nx < 17; nx++) WrZ80(R->IX.W + nx, dpb[nx]);
	ret_ok(R);
}




static void fid_seek(int drive, long track, long sector)
{
	long offs = track * 16384 + sector * 512 + 256;
	long clen;
	fseek(images[drive], 0, SEEK_END);	/* Get current size */
	clen = ftell(images[drive]);
	while (clen < offs)
	{
		fputc(0xE5, images[drive]);
		++clen;
	}
	fseek(images[drive], offs, SEEK_SET);
}



void fid_d_read(Z80 *R)	
{
	int nx;

	byte drv = R->BC.B.h;	/* (v1.22) BIOS floppy access support */

	lastdrv = drv;
	if (!open_image(drv)) 
	{
		ret_err(R);
		return;
	}
	if (drives[drv] >= 0) 
	{
		fid_fd_read(drives[drv], R);
		return;
	}
	
	fid_seek(R->BC.B.h, R->HL.W, R->DE.W);

	for (nx = 0; nx < 512; nx++) 
	{
		int cbyt = fgetc(images[R->BC.B.h]);
		if (cbyt == EOF) cbyt = 0xE5;
		WrZ80(R->IY.W++, cbyt);
	}
	ret_ok(R);
}


void fid_d_mess(Z80 *R)	/* (v1.22) expand disc error message */
{
	char msg[94];

	sprintf(msg,"%c: %s", lastdrv + 'A', fid_fd_mess(R));

	R->HL.W = fidtext;
	R->AF.B.l |= C_FLAG;
	fid_puts1(fidtext, msg);			
}


void fid_d_write(Z80 *R)
{
	int nx;
	byte drv = R->BC.B.h;	/* (v1.22) BIOS floppy access support */

	lastdrv = drv;
	if (!open_image(drv)) 
	{
		ret_err(R);
		return;
	}
	if (drives[drv] >= 0) 
	{
		fid_fd_read(drives[drv], R);
		return;
	}

	fid_seek(R->BC.B.h, R->HL.W, R->DE.W);

	for (nx = 0; nx < 512; nx++) 
	{
		if (fputc(RdZ80(R->IY.W++), images[R->BC.B.h]) == EOF)
		{
			ret_err(R);
			return;
		}
	}
	ret_ok(R);
}


void fid_d_flush(Z80 *R)
{
	byte drv = R->BC.B.h;

	lastdrv = drv;
	if (drives[drv] >= 0) 	/* (v1.22) BIOS floppy access support */
	{
		fid_fd_flush(drives[drv], R);
		return;
	}

	if (!open_image(drv)) 
	close_image(R->BC.B.h);
	ret_ok(R);
}


void setup_pars(Z80 *R)		/* v1.22: Determine whether a drive is floppy */
{				/* or hard, and set HL IX IY accordingly */
	int img = R->BC.B.h;
	char *dname;

	dskfil[0]  = img + 'a';
	dskfilu[0] = img + 'A';
	dname = getenv(dskfil);
	if (!dname) dname = getenv(dskfilu);
	if (!dname)
	{
		dname = "-----";
	}
	if (allow_floppy && isdisk(dname))	/* Floppy */
	{
		R->HL.W = 91;
		R->IX.W = 64;
		R->IY.W = 1024;
	}
	else
	{
		R->HL.W = 513;
		R->IX.W = 0;
		R->IY.W = 0;
	}
}



void fid_ems(Z80 *R)
{
	static int nhooked;
	static byte cdrv;
	char str[90];

/* This is called with: 

	A = 7
	B = flags, currently 0
	C = drive letter, currently 0FFh
	D = error:  0 => a drive has been hooked, E is its letter
		    1 => hook failure, no more memory
		    2 => hook failure, drive letter in use
		    3 => version control error
		    4 => start of fid_ems routine
	HL -> 93-byte message buffer

    returns: A=0 => ret NC
	     A=1 => ret C
	     A=2 => attempt to hook another disc drive
		      B=new drive letter

*/

	switch (R->DE.B.h)
	{
		case 4:			/* Start of FIDEMS */
		if (R->BC.B.l != 0xFF) allow_floppy = 1;
		hooked[0] = 0;	/* No drives hooked yet */
		nhooked = 0;
		cdrv = 2;	/* Start trying to hook at C: */
		R->BC.B.h = cdrv;
		R->AF.B.h = 2;
		fidtext = R->HL.W;	/* (v1.22) save text buffer */
		setup_pars(R);
		return;
	
		case 3:			/* Version control */
		fid_puts(R->HL.W, "JOYCE driver: Incorrect FID environment");
		R->AF.B.h = 0;
		return;

		case 2:			/* Drive letter in use */
		if (cdrv < 0x0f)	/* (v1.22) Failed to get C: - O: */
		{			/* This is because that */
					/* letter's in use, so try the */
			R->AF.B.h = 2;	/* next letter */
			++cdrv;
			R->BC.B.h = cdrv;
			setup_pars(R);
		}
		else if (nhooked)
		{
			sprintf(str,"JOYCE driver on drives %s", hooked);
			fid_puts(R->HL.W,str);
			R->AF.B.h = 1;
		}
		else
		{
			fid_puts(R->HL.W, "JOYCE driver: No free drive letters");
			R->AF.B.h = 0;
		}
		return;
		case 1:			/* No free memory */
		if (nhooked)
		{
			sprintf(str,"JOYCE driver on drives %s", hooked);
			fid_puts(R->HL.W,str);
			R->AF.B.h = 1;
			return;
		}
		else
		{
			fid_puts(R->HL.W, "JOYCE driver: No memory available");
			R->AF.B.h = 0;
			return;
		}
		case 0:			/* Drive hooked OK */
		nhooked++;
		sprintf(str,"%c: ", R->DE.B.l + 'A'); 
		strcat(hooked, str);		
		R->AF.B.h = 2;
		++cdrv;
		R->BC.B.h = cdrv;
		setup_pars(R);
		return;
	}

}



void fid_cems(Z80 *R)
{
/* This is called with: 

	A = 11
	B = flags, currently 0
	C = 5
	D = error:  0 => a device has been got
		    1 => hook failure, no more devices
		    2 => hook failure, device name in use
		    3 => version control error
	HL -> 93-byte message buffer

    returns: A=0 => ret NC
	     A=1 => ret C
*/

	switch (R->DE.B.h)
	{
		case 3:			/* Version control */
		fid_puts(R->HL.W, "VGA.FID: Incorrect FID environment");
		R->AF.B.h = 0;
		return;

		case 2:			/* Drive letter in use */
		fid_puts(R->HL.W, "VGA.FID: VGA driver already loaded");
		R->AF.B.h = 0;
		return;

		case 1:			/* No free memory */
		fid_puts(R->HL.W, "VGA.FID: No character devices available");
		R->AF.B.h = 0;
		return;

		case 0:			/* Device hooked OK */
		fid_puts(R->HL.W, "VGA.FID loaded OK");
		R->AF.B.h = 1;
		return;
	}

}


void vga_capabilities(Z80 *R)	/* This function new in v1.20 */
{
	/* Sets the following bits in HL: */
	/* Bit 0: If 800x600 screen is available */
	/* Bit 1: If GSX is available */
	/* Bit 2: (v1.22) If LIOS is available */
	
#ifdef __MSDOS__
	R->HL.W = 7;
#else
	R->HL.W = 5;	/* No GSX.  */
#endif
	R->DE.W = 0;
	R->BC.W = 0;
	R->IX.W = 0;
	R->IY.W = 0;
}



void fid_char(Z80 *R)
{
	switch(R->BC.B.l)
	{
		case 0: vga_setmode(0); break;
		case 1: vga_setmode(1); break;
		case 2: vga_putchar(R->DE.B.l); break;
		case 3: vga_putmch(R->DE.B.l);  break;
		case 4: gsx_cmd(R->DE.W); break;	/* v1.20 */
		case 5: fid_cems(R); break;
		case 6: vga_capabilities(R); break;	/* v1.20 */
	}
}

/* FIDEMS for the COM port driver */

void com_ems(Z80 *R)
{
	char str[90];

/* This is called with: 

	A = 0Dh
	C = 7
	D = error:  0 => a device has been got
		    1 => hook failure, no more devices
		    2 => hook failure, device name in use
		    3 => version control error
	HL -> 93-byte message buffer

    returns: A=0 => ret NC
	     A=1 => ret C
*/
	switch (R->DE.B.h)
	{
		case 3:			/* Version control */
		fid_puts(R->HL.W, "COMPORT.FID: Incorrect FID environment");
		R->AF.B.h = 0;
		return;

		case 2:			/* Device name in use*/
		fid_puts(R->HL.W, "COMPORT.FID: A COM port driver is already loaded");
		R->AF.B.h = 0;
		return;

		case 1:			/* No free memory */
		fid_puts(R->HL.W, "COMPORT.FID: No character devices available");
		R->AF.B.h = 0;
		return;

		case 0:			/* Device hooked OK */
		sprintf(str,"COMPORT.FID loaded. Using port COM%d", jset.comport + 1);
		fid_puts(R->HL.W, str);
		R->AF.B.h = 1;
		return;
	}
}
